"""
@author: Amine Sehili <amine.sehili@gmail.com>
"""

from shutil import which
from unittest.mock import patch

import numpy as np
import pytest

from auditok.io import (
    AudioIOError,
    AudioParameterError,
    BufferAudioSource,
    FFmpegAudioSource,
    RawAudioSource,
    WaveAudioSource,
    from_file,
)
from auditok.signal import SAMPLE_WIDTH_TO_DTYPE

requires_ffmpeg = pytest.mark.skipif(
    which("ffmpeg") is None, reason="ffmpeg not available"
)


def _sample_generator(*data_buffers):
    """
    Takes a list of many mono audio data buffers and makes a sample generator
    of interleaved audio samples, one sample from each channel. The resulting
    generator can be used to build a multichannel audio buffer.
    >>> gen = _sample_generator("abcd", "ABCD")
    >>> list(gen)
    ["a", "A", "b", "B", "c", "C", "d", "D"]
    """
    frame_gen = zip(*data_buffers)
    return (sample for frame in frame_gen for sample in frame)


def _generate_pure_tone(
    frequency, duration_sec=1, sampling_rate=16000, sample_width=2, volume=1e4
):
    """
    Generates a pure tone with the given frequency.
    """
    assert frequency <= sampling_rate / 2
    max_value = (2 ** (sample_width * 8) // 2) - 1
    if volume > max_value:
        volume = max_value
    dtype = SAMPLE_WIDTH_TO_DTYPE[sample_width]
    total_samples = int(sampling_rate * duration_sec)
    step = frequency / sampling_rate
    two_pi_step = 2 * np.pi * step
    data = np.array(
        [int(np.sin(two_pi_step * i) * volume) for i in range(total_samples)]
    ).astype(dtype)
    return data


@pytest.fixture
def pure_tone_data(freq):

    PURE_TONE_DICT = {
        freq: _generate_pure_tone(freq, 1, 16000, 2)
        for freq in (400, 800, 1600)
    }
    PURE_TONE_DICT.update(
        {
            freq: _generate_pure_tone(freq, 0.1, 16000, 2)
            for freq in (600, 1150, 2400, 7220)
        }
    )
    return PURE_TONE_DICT[freq]


PURE_TONE_DICT = {
    freq: _generate_pure_tone(freq, 1, 16000, 2) for freq in (400, 800, 1600)
}
PURE_TONE_DICT.update(
    {
        freq: _generate_pure_tone(freq, 0.1, 16000, 2)
        for freq in (600, 1150, 2400, 7220)
    }
)


def audio_source_read_all_gen(audio_source, size=None):
    if size is None:
        size = int(audio_source.sr * 0.1)  # 100ms
    while True:
        data = audio_source.read(size)
        if data is None:
            break
        yield data


@pytest.mark.parametrize(
    "file_suffix, frequencies",
    [
        ("mono_400Hz", (400,)),  # mono
        ("3channel_400-800-1600Hz", (400, 800, 1600)),  # multichannel
    ],
    ids=["mono", "multichannel"],
)
def test_BufferAudioSource_read_all(file_suffix, frequencies):
    file = "tests/data/test_16KHZ_{}.raw".format(file_suffix)
    with open(file, "rb") as fp:
        expected = fp.read()
    channels = len(frequencies)
    audio_source = BufferAudioSource(expected, 16000, 2, channels)
    audio_source.open()
    data = audio_source.read(None)
    assert data == expected
    audio_source.rewind()
    data = audio_source.read(-10)
    assert data == expected
    audio_source.close()


@pytest.mark.parametrize(
    "file_suffix, frequencies",
    [
        ("mono_400Hz", (400,)),  # mono
        ("3channel_400-800-1600Hz", (400, 800, 1600)),  # multichannel
    ],
    ids=["mono", "multichannel"],
)
def test_RawAudioSource(file_suffix, frequencies):
    file = "tests/data/test_16KHZ_{}.raw".format(file_suffix)
    channels = len(frequencies)
    audio_source = RawAudioSource(file, 16000, 2, channels)
    audio_source.open()
    data_read_all = b"".join(audio_source_read_all_gen(audio_source))
    audio_source.close()
    mono_channels = [PURE_TONE_DICT[freq] for freq in frequencies]
    dtype = SAMPLE_WIDTH_TO_DTYPE[audio_source.sample_width]
    expected = np.fromiter(_sample_generator(*mono_channels), dtype).tobytes()

    assert data_read_all == expected

    # assert read all data with None
    audio_source = RawAudioSource(file, 16000, 2, channels)
    audio_source.open()
    data_read_all = audio_source.read(None)
    audio_source.close()
    assert data_read_all == expected

    # assert read all data with a negative size
    audio_source = RawAudioSource(file, 16000, 2, channels)
    audio_source.open()
    data_read_all = audio_source.read(-10)
    audio_source.close()
    assert data_read_all == expected


@pytest.mark.parametrize(
    "file_suffix, frequencies",
    [
        ("mono_400Hz", (400,)),  # mono
        ("3channel_400-800-1600Hz", (400, 800, 1600)),  # multichannel
    ],
    ids=["mono", "multichannel"],
)
def test_WaveAudioSource(file_suffix, frequencies):
    file = "tests/data/test_16KHZ_{}.wav".format(file_suffix)
    audio_source = WaveAudioSource(file)
    audio_source.open()
    data = b"".join(audio_source_read_all_gen(audio_source))
    audio_source.close()
    mono_channels = [PURE_TONE_DICT[freq] for freq in frequencies]
    dtype = SAMPLE_WIDTH_TO_DTYPE[audio_source.sample_width]
    expected = np.fromiter(_sample_generator(*mono_channels), dtype).tobytes()

    assert data == expected

    # assert read all data with None
    audio_source = WaveAudioSource(file)
    audio_source.open()
    data_read_all = audio_source.read(None)
    audio_source.close()
    assert data_read_all == expected

    # assert read all data with a negative size
    audio_source = WaveAudioSource(file)
    audio_source.open()
    data_read_all = audio_source.read(-10)
    audio_source.close()
    assert data_read_all == expected


@requires_ffmpeg
@pytest.mark.parametrize(
    "filename, sr, ch",
    [
        ("DTMF_tones_44.1KHZ_stereo.flac", 44100, 2),
        ("DTMF_tones_44.1KHZ_stereo.ogg", 44100, 2),
        ("DTMF_tones_44.1KHZ_stereo.mp3", 44100, 2),
        ("DTMF_tones_16KHZ_mono.flac", 16000, 1),
        ("DTMF_tones_16KHZ_mono.ogg", 16000, 1),
        ("DTMF_tones_16KHZ_mono.mp3", 16000, 1),
    ],
    ids=[
        "stereo_44.1KHZ_flac",
        "stereo_44.1KHZ_ogg",
        "stereo_44.1KHZ_mp3",
        "mono_16KHZ_flac",
        "mono_16KHZ_ogg",
        "mono_16KHZ_mp3",
    ],
)
def test_FFmpegAudioSource(filename, sr, ch):
    test_file = "tests/data/{}".format(filename)
    audio_source = FFmpegAudioSource(test_file)
    assert audio_source.sr == sr
    assert audio_source.sw == 2
    assert audio_source.ch == ch

    # read all with None
    data = audio_source.read(None)
    audio_source.close()
    expected_length = sr * 2 * ch  # 1s * sw * ch
    assert len(data) == expected_length

    # read incrementally
    audio_source = FFmpegAudioSource(test_file)
    data_incremental = b"".join(audio_source_read_all_gen(audio_source))
    audio_source.close()
    assert data_incremental == data


@requires_ffmpeg
def test_FFmpegAudioSource_24bit_flac_outputs_16bit():
    test_file = "tests/data/DTMF_tones_44.1KHZ_stereo_24bit.flac"
    audio_source = FFmpegAudioSource(test_file)
    assert audio_source.sr == 44100
    assert audio_source.sw == 2
    assert audio_source.ch == 2
    data = audio_source.read(None)
    audio_source.close()
    expected_length = 44100 * 2 * 2  # 1s * sw * ch
    assert len(data) == expected_length


@requires_ffmpeg
def test_FFmpegAudioSource_resample():
    test_file = "tests/data/DTMF_tones_44.1KHZ_stereo.flac"
    audio_source = FFmpegAudioSource(test_file, sampling_rate=16000)
    assert audio_source.sr == 16000
    assert audio_source.sw == 2
    assert audio_source.ch == 2
    data = audio_source.read(None)
    audio_source.close()
    expected_length = 16000 * 2 * 2  # 1s * sw * ch
    assert len(data) == expected_length


@requires_ffmpeg
def test_FFmpegAudioSource_downmix_to_mono():
    test_file = "tests/data/DTMF_tones_44.1KHZ_stereo.flac"
    audio_source = FFmpegAudioSource(test_file, channels=1)
    assert audio_source.sr == 44100
    assert audio_source.sw == 2
    assert audio_source.ch == 1
    data = audio_source.read(None)
    audio_source.close()
    expected_length = 44100 * 2 * 1  # 1s * sw * ch
    assert len(data) == expected_length


@requires_ffmpeg
def test_FFmpegAudioSource_change_sample_width():
    test_file = "tests/data/DTMF_tones_16KHZ_mono.flac"
    audio_source = FFmpegAudioSource(test_file, sample_width=4)
    assert audio_source.sr == 16000
    assert audio_source.sw == 4
    assert audio_source.ch == 1
    data = audio_source.read(None)
    audio_source.close()
    expected_length = 16000 * 4 * 1  # 1s * sw * ch
    assert len(data) == expected_length


@requires_ffmpeg
def test_FFmpegAudioSource_convert_all_params():
    test_file = "tests/data/DTMF_tones_44.1KHZ_stereo.flac"
    audio_source = FFmpegAudioSource(
        test_file, sampling_rate=16000, sample_width=1, channels=1
    )
    assert audio_source.sr == 16000
    assert audio_source.sw == 1
    assert audio_source.ch == 1
    data = audio_source.read(None)
    audio_source.close()
    expected_length = 16000 * 1 * 1  # 1s * sw * ch
    assert len(data) == expected_length


@requires_ffmpeg
def test_FFmpegAudioSource_invalid_sample_width():
    test_file = "tests/data/DTMF_tones_16KHZ_mono.flac"
    with pytest.raises(AudioParameterError, match="sample_width must be"):
        FFmpegAudioSource(test_file, sample_width=3)


@requires_ffmpeg
def test_from_file_with_conversion_params():
    test_file = "tests/data/DTMF_tones_44.1KHZ_stereo.flac"
    source = from_file(test_file, sr=16000, ch=1)
    assert source.sr == 16000
    assert source.sw == 2
    assert source.ch == 1
    data = source.read(None)
    source.close()
    expected_length = 16000 * 2 * 1
    assert len(data) == expected_length


def test_from_file_wav_does_not_use_FFmpegAudioSource():
    wav_file = "tests/data/DTMF_tones_44.1KHZ_mono.wav"
    with patch("auditok.io.FFmpegAudioSource") as mock_ffmpeg:
        source = from_file(wav_file)
        assert not mock_ffmpeg.called
    assert isinstance(source, BufferAudioSource)


class TestBufferAudioSource_SR10_SW1_CH1:
    @pytest.fixture(autouse=True)
    def setup_and_teardown(self):
        self.data = b"ABCDEFGHIJKLMNOPQRSTUVWXYZ012345"
        self.audio_source = BufferAudioSource(
            data=self.data, sampling_rate=10, sample_width=1, channels=1
        )
        self.audio_source.open()
        yield
        self.audio_source.close()

    def test_sr10_sw1_ch1_read_1(self):
        block = self.audio_source.read(1)
        exp = b"A"
        assert block == exp

    def test_sr10_sw1_ch1_read_6(self):
        block = self.audio_source.read(6)
        exp = b"ABCDEF"
        assert block == exp

    def test_sr10_sw1_ch1_read_multiple(self):
        block = self.audio_source.read(1)
        exp = b"A"
        assert block == exp

        block = self.audio_source.read(6)
        exp = b"BCDEFG"
        assert block == exp

        block = self.audio_source.read(13)
        exp = b"HIJKLMNOPQRST"
        assert block == exp

        block = self.audio_source.read(9999)
        exp = b"UVWXYZ012345"
        assert block == exp

    def test_sr10_sw1_ch1_read_all(self):
        block = self.audio_source.read(9999)
        assert block == self.data

        block = self.audio_source.read(1)
        assert block is None

    def test_sr10_sw1_ch1_sampling_rate(self):
        srate = self.audio_source.sampling_rate
        assert srate == 10

    def test_sr10_sw1_ch1_sample_width(self):
        swidth = self.audio_source.sample_width
        assert swidth == 1

    def test_sr10_sw1_ch1_channels(self):
        channels = self.audio_source.channels
        assert channels == 1

    @pytest.mark.parametrize(
        "block_sizes, expected_sample, expected_second, expected_ms",
        [
            ([], 0, 0, 0),  # empty
            ([0], 0, 0, 0),  # zero
            ([5], 5, 0.5, 500),  # five
            ([5, 20], 25, 2.5, 2500),  # multiple
        ],
        ids=["empty", "zero", "five", "multiple"],
    )
    def test_position(
        self, block_sizes, expected_sample, expected_second, expected_ms
    ):
        for block_size in block_sizes:
            self.audio_source.read(block_size)
        position = self.audio_source.position
        assert position == expected_sample

        position_s = self.audio_source.position_s
        assert position_s == expected_second

        position_ms = self.audio_source.position_ms
        assert position_ms == expected_ms

    @pytest.mark.parametrize(
        "position, expected_sample, expected_second, expected_ms",
        [
            (0, 0, 0, 0),  # zero
            (1, 1, 0.1, 100),  # one
            (10, 10, 1, 1000),  # ten
            (-1, 31, 3.1, 3100),  # negative_1
            (-7, 25, 2.5, 2500),  # negative_2
        ],
        ids=["zero", "one", "ten", "negative_1", "negative_2"],
    )
    def test_position_setter(
        self, position, expected_sample, expected_second, expected_ms
    ):
        self.audio_source.position = position

        position = self.audio_source.position
        assert position == expected_sample

        position_s = self.audio_source.position_s
        assert position_s == expected_second

        position_ms = self.audio_source.position_ms
        assert position_ms == expected_ms

    @pytest.mark.parametrize(
        "position_s, expected_sample, expected_second, expected_ms",
        [
            (0, 0, 0, 0),  # zero
            (0.1, 1, 0.1, 100),  # one
            (1, 10, 1, 1000),  # ten
            (-0.1, 31, 3.1, 3100),  # negative_1
            (-0.7, 25, 2.5, 2500),  # negative_2
        ],
        ids=["zero", "one", "ten", "negative_1", "negative_2"],
    )
    def test_position_s_setter(
        self, position_s, expected_sample, expected_second, expected_ms
    ):
        self.audio_source.position_s = position_s

        position = self.audio_source.position
        assert position == expected_sample

        position_s = self.audio_source.position_s
        assert position_s == expected_second

        position_ms = self.audio_source.position_ms
        assert position_ms == expected_ms

    @pytest.mark.parametrize(
        "position_ms, expected_sample, expected_second, expected_ms",
        [
            (0, 0, 0, 0),  # zero
            (100, 1, 0.1, 100),  # one
            (1000, 10, 1, 1000),  # ten
            (-100, 31, 3.1, 3100),  # negative_1
            (-700, 25, 2.5, 2500),  # negative_2
        ],
        ids=["zero", "one", "ten", "negative_1", "negative_2"],
    )
    def test_position_ms_setter(
        self, position_ms, expected_sample, expected_second, expected_ms
    ):
        self.audio_source.position_ms = position_ms

        position = self.audio_source.position
        assert position == expected_sample

        position_s = self.audio_source.position_s
        assert position_s == expected_second

        position_ms = self.audio_source.position_ms
        assert position_ms == expected_ms

    @pytest.mark.parametrize(
        "position",
        [
            100,  # positive
            -100,  # negative
        ],
        ids=["positive", "negative"],
    )
    def test_position_setter_out_of_range(self, position):
        with pytest.raises(IndexError):
            self.audio_source.position = position

    @pytest.mark.parametrize(
        "position_s",
        [
            100,  # positive
            -100,  # negative
        ],
        ids=["positive", "negative"],
    )
    def test_position_s_setter_out_of_range(self, position_s):
        with pytest.raises(IndexError):
            self.audio_source.position_s = position_s

    @pytest.mark.parametrize(
        "position_ms",
        [
            10000,  # positive
            -10000,  # negative
        ],
        ids=["positive", "negative"],
    )
    def test_position_ms_setter_out_of_range(self, position_ms):
        with pytest.raises(IndexError):
            self.audio_source.position_ms = position_ms

    def test_sr10_sw1_ch1_initial_position_s_0(self):
        tp = self.audio_source.position_s
        assert tp == 0.0

    def test_sr10_sw1_ch1_position_s_1_after_read(self):
        srate = self.audio_source.sampling_rate
        # read one second
        self.audio_source.read(srate)
        tp = self.audio_source.position_s
        assert tp == 1.0

    def test_sr10_sw1_ch1_position_s_2_5(self):
        # read 2.5 seconds
        self.audio_source.read(25)
        tp = self.audio_source.position_s
        assert tp == 2.5

    def test_sr10_sw1_ch1_position_s_0(self):
        self.audio_source.read(10)
        self.audio_source.position_s = 0
        tp = self.audio_source.position_s
        assert tp == 0.0

    def test_sr10_sw1_ch1_position_s_1(self):
        self.audio_source.position_s = 1
        tp = self.audio_source.position_s
        assert tp == 1.0

    def test_sr10_sw1_ch1_rewind(self):
        self.audio_source.read(10)
        self.audio_source.rewind()
        tp = self.audio_source.position
        assert tp == 0

    def test_sr10_sw1_ch1_read_closed(self):
        self.audio_source.close()
        with pytest.raises(AudioIOError):
            self.audio_source.read(1)


class TestBufferAudioSource_SR16_SW2_CH1:
    @pytest.fixture(autouse=True)
    def setup_and_teardown(self):
        self.data = b"ABCDEFGHIJKLMNOPQRSTUVWXYZ012345"
        self.audio_source = BufferAudioSource(
            data=self.data, sampling_rate=16, sample_width=2, channels=1
        )
        self.audio_source.open()
        yield
        self.audio_source.close()

    def test_sr16_sw2_ch1_read_1(self):
        block = self.audio_source.read(1)
        exp = b"AB"
        assert block == exp

    def test_sr16_sw2_ch1_read_6(self):
        block = self.audio_source.read(6)
        exp = b"ABCDEFGHIJKL"
        assert block == exp

    def test_sr16_sw2_ch1_read_multiple(self):
        block = self.audio_source.read(1)
        exp = b"AB"
        assert block == exp

        block = self.audio_source.read(6)
        exp = b"CDEFGHIJKLMN"
        assert block == exp

        block = self.audio_source.read(5)
        exp = b"OPQRSTUVWX"
        assert block == exp

        block = self.audio_source.read(9999)
        exp = b"YZ012345"
        assert block == exp

    def test_sr16_sw2_ch1_read_all(self):
        block = self.audio_source.read(9999)
        assert block == self.data

        block = self.audio_source.read(1)
        assert block is None

    def test_sr16_sw2_ch1_sampling_rate(self):
        srate = self.audio_source.sampling_rate
        assert srate == 16

    def test_sr16_sw2_ch1_sample_width(self):
        swidth = self.audio_source.sample_width
        assert swidth == 2

    def test_sr16_sw2_ch1_channels(self):
        channels = self.audio_source.channels
        assert channels == 1

    @pytest.mark.parametrize(
        "block_sizes, expected_sample, expected_second, expected_ms",
        [
            ([], 0, 0, 0),  # empty
            ([0], 0, 0, 0),  # zero
            ([2], 2, 2 / 16, int(2000 / 16)),  # two
            ([11], 11, 11 / 16, int(11 * 1000 / 16)),  # eleven
            ([4, 8], 12, 0.75, 750),  # multiple
        ],
        ids=["empty", "zero", "two", "eleven", "multiple"],
    )
    def test_position(
        self, block_sizes, expected_sample, expected_second, expected_ms
    ):
        for block_size in block_sizes:
            self.audio_source.read(block_size)
        position = self.audio_source.position
        assert position == expected_sample

        position_s = self.audio_source.position_s
        assert position_s == expected_second

        position_ms = self.audio_source.position_ms
        assert position_ms == expected_ms

    def test_sr16_sw2_ch1_read_position_0(self):
        self.audio_source.read(10)
        self.audio_source.position = 0
        pos = self.audio_source.position
        assert pos == 0

    @pytest.mark.parametrize(
        "position, expected_sample, expected_second, expected_ms",
        [
            (0, 0, 0, 0),  # zero
            (1, 1, 1 / 16, int(1000 / 16)),  # one
            (10, 10, 10 / 16, int(10000 / 16)),  # ten
            (-1, 15, 15 / 16, int(15000 / 16)),  # negative_1
            (-7, 9, 9 / 16, int(9000 / 16)),  # negative_2
        ],
        ids=["zero", "one", "ten", "negative_1", "negative_2"],
    )
    def test_position_setter(
        self, position, expected_sample, expected_second, expected_ms
    ):
        self.audio_source.position = position

        position = self.audio_source.position
        assert position == expected_sample

        position_s = self.audio_source.position_s
        assert position_s == expected_second

        position_ms = self.audio_source.position_ms
        assert position_ms == expected_ms

    @pytest.mark.parametrize(
        "position_s, expected_sample, expected_second, expected_ms",
        [
            (0, 0, 0, 0),  # zero
            (0.1, 1, 1 / 16, int(1000 / 16)),  # one
            (1 / 8, 2, 1 / 8, int(1 / 8 * 1000)),  # two
            (0.75, 12, 0.75, 750),  # twelve
            (-0.1, 15, 15 / 16, int(15000 / 16)),  # negative_1
            (-0.7, 5, 5 / 16, int(5000 / 16)),  # negative_2
        ],
        ids=["zero", "one", "two", "twelve", "negative_1", "negative_2"],
    )
    def test_position_s_setter(
        self, position_s, expected_sample, expected_second, expected_ms
    ):
        self.audio_source.position_s = position_s

        position = self.audio_source.position
        assert position == expected_sample

        position_s = self.audio_source.position_s
        assert position_s == expected_second

        position_ms = self.audio_source.position_ms
        assert position_ms == expected_ms

    @pytest.mark.parametrize(
        "position_ms, expected_sample, expected_second, expected_ms",
        [
            (0, 0, 0, 0),  # zero
            (100, 1, 1 / 16, int(1000 / 16)),  # one
            (1000, 16, 1, 1000),  # ten
            (-100, 15, 15 / 16, int(15 * 1000 / 16)),  # negative_1
            (-500, 8, 0.5, 500),  # negative_2
            (-700, 5, 5 / 16, int(5 * 1000 / 16)),  # negative_3
        ],
        ids=["zero", "one", "ten", "negative_1", "negative_2", "negative_3"],
    )
    def test_position_ms_setter(
        self, position_ms, expected_sample, expected_second, expected_ms
    ):
        self.audio_source.position_ms = position_ms

        position = self.audio_source.position
        assert position == expected_sample

        position_s = self.audio_source.position_s
        assert position_s == expected_second

        position_ms = self.audio_source.position_ms
        assert position_ms == expected_ms

    def test_sr16_sw2_ch1_rewind(self):
        self.audio_source.read(10)
        self.audio_source.rewind()
        tp = self.audio_source.position
        assert tp == 0


class TestBufferAudioSource_SR11_SW4_CH1:
    @pytest.fixture(autouse=True)
    def setup_and_teardown(self):
        self.data = b"ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789abcdefgh"
        self.audio_source = BufferAudioSource(
            data=self.data, sampling_rate=11, sample_width=4, channels=1
        )
        self.audio_source.open()
        yield
        self.audio_source.close()

    def test_sr11_sw4_ch1_read_1(self):
        block = self.audio_source.read(1)
        exp = b"ABCD"
        assert block == exp

    def test_sr11_sw4_ch1_read_6(self):
        block = self.audio_source.read(6)
        exp = b"ABCDEFGHIJKLMNOPQRSTUVWX"
        assert block == exp

    def test_sr11_sw4_ch1_read_multiple(self):
        block = self.audio_source.read(1)
        exp = b"ABCD"
        assert block == exp

        block = self.audio_source.read(6)
        exp = b"EFGHIJKLMNOPQRSTUVWXYZ01"
        assert block == exp

        block = self.audio_source.read(3)
        exp = b"23456789abcd"
        assert block == exp

        block = self.audio_source.read(9999)
        exp = b"efgh"
        assert block == exp

    def test_sr11_sw4_ch1_read_all(self):
        block = self.audio_source.read(9999)
        assert block == self.data

        block = self.audio_source.read(1)
        assert block is None

    def test_sr11_sw4_ch1_sampling_rate(self):
        srate = self.audio_source.sampling_rate
        assert srate == 11

    def test_sr11_sw4_ch1_sample_width(self):
        swidth = self.audio_source.sample_width
        assert swidth == 4

    def test_sr11_sw4_ch1_channels(self):
        channels = self.audio_source.channels
        assert channels == 1

    def test_sr11_sw4_ch1_intial_position_0(self):
        pos = self.audio_source.position
        assert pos == 0

    def test_sr11_sw4_ch1_position_5(self):
        self.audio_source.read(5)
        pos = self.audio_source.position
        assert pos == 5

    def test_sr11_sw4_ch1_position_9(self):
        self.audio_source.read(5)
        self.audio_source.read(4)
        pos = self.audio_source.position
        assert pos == 9

    def test_sr11_sw4_ch1_position_0(self):
        self.audio_source.read(10)
        self.audio_source.position = 0
        pos = self.audio_source.position
        assert pos == 0

    def test_sr11_sw4_ch1_position_10(self):
        self.audio_source.position = 10
        pos = self.audio_source.position
        assert pos == 10

    def test_sr11_sw4_ch1_initial_position_s_0(self):
        tp = self.audio_source.position_s
        assert tp == 0.0

    def test_sr11_sw4_ch1_position_s_1_after_read(self):
        srate = self.audio_source.sampling_rate
        # read one second
        self.audio_source.read(srate)
        tp = self.audio_source.position_s
        assert tp == 1.0

    def test_sr11_sw4_ch1_position_s_0_63(self):
        # read 2.5 seconds
        self.audio_source.read(7)
        tp = self.audio_source.position_s
        assert tp, pytest.approx(0.636363636364)

    def test_sr11_sw4_ch1_position_s_0(self):
        self.audio_source.read(10)
        self.audio_source.position_s = 0
        tp = self.audio_source.position_s
        assert tp == 0.0

    def test_sr11_sw4_ch1_position_s_1(self):
        self.audio_source.position_s = 1
        tp = self.audio_source.position_s
        assert tp == 1.0

    def test_sr11_sw4_ch1_rewind(self):
        self.audio_source.read(10)
        self.audio_source.rewind()
        tp = self.audio_source.position
        assert tp == 0


class TestBufferAudioSourceCreationException:
    def test_wrong_sample_width_value(self):
        with pytest.raises(AudioParameterError) as audio_param_err:
            _ = BufferAudioSource(
                data=b"ABCDEFGHI", sampling_rate=9, sample_width=3, channels=1
            )
        assert (
            str(audio_param_err.value)
            == "Sample width must be one of: 1, 2 or 4 (bytes)"
        )

    def test_wrong_data_buffer_size(self):
        with pytest.raises(AudioParameterError) as audio_param_err:
            _ = BufferAudioSource(
                data=b"ABCDEFGHI", sampling_rate=8, sample_width=2, channels=1
            )
        msg = "The length of audio data must be an integer multiple of "
        msg += "`sample_width * channels`"
        assert str(audio_param_err.value) == msg


class TestAudioSourceProperties:
    def test_read_properties(self):
        data = b""
        sampling_rate = 8000
        sample_width = 2
        channels = 1
        a_source = BufferAudioSource(
            data, sampling_rate, sample_width, channels
        )

        assert a_source.sampling_rate == sampling_rate
        assert a_source.sample_width == sample_width
        assert a_source.channels == channels

    def test_set_readonly_properties_exception(self):
        data = b""
        sampling_rate = 8000
        sample_width = 2
        channels = 1
        a_source = BufferAudioSource(
            data, sampling_rate, sample_width, channels
        )

        with pytest.raises(AttributeError):
            a_source.sampling_rate = 16000

        with pytest.raises(AttributeError):
            a_source.sample_width = 1

        with pytest.raises(AttributeError):
            a_source.channels = 2


class TestAudioSourceShortProperties:
    def test_read_short_properties(self):
        data = b""
        sampling_rate = 8000
        sample_width = 2
        channels = 1
        a_source = BufferAudioSource(
            data, sampling_rate, sample_width, channels
        )

        assert a_source.sr == sampling_rate
        assert a_source.sw == sample_width
        assert a_source.ch == channels

    def test_set_readonly_short_properties_exception(self):
        data = b""
        sampling_rate = 8000
        sample_width = 2
        channels = 1
        a_source = BufferAudioSource(
            data, sampling_rate, sample_width, channels
        )

        with pytest.raises(AttributeError):
            a_source.sr = 16000

        with pytest.raises(AttributeError):
            a_source.sw = 1

        with pytest.raises(AttributeError):
            a_source.ch = 2



import os
from tempfile import TemporaryDirectory
from unittest.mock import Mock, call, patch

import pytest

import auditok.workers
from auditok import AudioReader, AudioRegion, split, split_and_join_with_silence
from auditok.cmdline_util import make_logger
from auditok.workers import (
    AudioEventsJoinerWorker,
    CommandLineWorker,
    PlayerWorker,
    PrintWorker,
    RegionSaverWorker,
    StreamSaverWorker,
    TokenizerWorker,
)


@pytest.fixture
def audio_data_source():
    reader = AudioReader(
        input="tests/data/test_split_10HZ_mono.raw",
        block_dur=0.1,
        sr=10,
        sw=2,
        ch=1,
    )
    yield reader
    reader.close()


@pytest.fixture
def expected_detections():
    return [
        (0.2, 1.6),
        (1.7, 3.1),
        (3.4, 5.4),
        (5.4, 7.4),
        (7.4, 7.6),
    ]


def test_TokenizerWorker(audio_data_source, expected_detections):
    with TemporaryDirectory() as tmpdir:
        file = os.path.join(tmpdir, "file.log")
        logger = make_logger(file=file, name="test_TokenizerWorker")
        tokenizer = TokenizerWorker(
            audio_data_source,
            logger=logger,
            min_dur=0.3,
            max_dur=2,
            max_silence=0.2,
            strict_min_dur=False,
            eth=50,
        )
        tokenizer.start_all()
        tokenizer.join()
        with open(file) as fp:
            log_lines = fp.readlines()

    log_fmt = (
        "[DET]: Detection {} (start: {:.3f}, end: {:.3f}, duration: {:.3f})"
    )
    assert len(tokenizer.detections) == len(expected_detections)
    for i, (det, exp, log_line) in enumerate(
        zip(
            tokenizer.detections,
            expected_detections,
            log_lines,
        ),
        1,
    ):
        start, end = exp
        exp_log_line = log_fmt.format(i, start, end, end - start)
        assert pytest.approx(det.start) == start
        assert pytest.approx(det.end) == end
        assert log_line[28:].strip() == exp_log_line


def test_PlayerWorker(audio_data_source, expected_detections):
    with TemporaryDirectory() as tmpdir:
        file = os.path.join(tmpdir, "file.log")
        logger = make_logger(file=file, name="test_RegionSaverWorker")
        player_mock = Mock()
        observers = [PlayerWorker(player_mock, logger=logger)]
        tokenizer = TokenizerWorker(
            audio_data_source,
            logger=logger,
            observers=observers,
            min_dur=0.3,
            max_dur=2,
            max_silence=0.2,
            strict_min_dur=False,
            eth=50,
        )
        tokenizer.start_all()
        tokenizer.join()
        tokenizer._observers[0].join()
        with open(file) as fp:
            log_lines = [
                line for line in fp.readlines() if line.startswith("[PLAY]")
            ]

    assert player_mock.play.called
    assert len(tokenizer.detections) == len(expected_detections)
    log_fmt = "[PLAY]: Detection {id} played"
    for i, (det, exp, log_line) in enumerate(
        zip(  # noqa: B905
            tokenizer.detections,
            expected_detections,
            log_lines,
        ),
        1,
    ):
        start, end = exp
        exp_log_line = log_fmt.format(id=i)
        assert pytest.approx(det.start) == start
        assert pytest.approx(det.end) == end
        assert log_line[28:].strip() == exp_log_line


def test_RegionSaverWorker(audio_data_source, expected_detections):
    filename_format = "Region_{id}_{start:.6f}-{end:.3f}_{duration:.3f}.wav"
    with TemporaryDirectory() as tmpdir:
        file = os.path.join(tmpdir, "file.log")
        logger = make_logger(file=file, name="test_RegionSaverWorker")
        observers = [RegionSaverWorker(filename_format, logger=logger)]
        tokenizer = TokenizerWorker(
            audio_data_source,
            logger=logger,
            observers=observers,
            min_dur=0.3,
            max_dur=2,
            max_silence=0.2,
            strict_min_dur=False,
            eth=50,
        )
        with patch("auditok.audio.AudioRegion.save") as patched_save:
            tokenizer.start_all()
            tokenizer.join()
            tokenizer._observers[0].join()
        with open(file) as fp:
            log_lines = [
                line for line in fp.readlines() if line.startswith("[SAVE]")
            ]

    expected_save_calls = [
        call(
            filename_format.format(
                id=i, start=exp[0], end=exp[1], duration=exp[1] - exp[0]
            ),
            None,
        )
        for i, exp in enumerate(expected_detections, 1)
    ]

    mock_calls = [
        c for i, c in enumerate(patched_save.mock_calls) if i % 2 == 0
    ]
    assert mock_calls == expected_save_calls
    assert len(tokenizer.detections) == len(expected_detections)

    log_fmt = "[SAVE]: Detection {id} saved as '{filename}'"
    for i, (det, exp, log_line) in enumerate(
        zip(  # noqa: B905
            tokenizer.detections,
            expected_detections,
            log_lines,
        ),
        1,
    ):
        start, end = exp
        expected_filename = filename_format.format(
            id=i, start=start, end=end, duration=end - start
        )
        exp_log_line = log_fmt.format(id=i, filename=expected_filename)
        assert pytest.approx(det.start) == start
        assert pytest.approx(det.end) == end
        assert log_line[28:].strip() == exp_log_line


def test_CommandLineWorker(audio_data_source, expected_detections):
    command_format = "do nothing with"
    with TemporaryDirectory() as tmpdir:
        file = os.path.join(tmpdir, "file.log")
        logger = make_logger(file=file, name="test_CommandLineWorker")
        observers = [CommandLineWorker(command_format, logger=logger)]
        tokenizer = TokenizerWorker(
            audio_data_source,
            logger=logger,
            observers=observers,
            min_dur=0.3,
            max_dur=2,
            max_silence=0.2,
            strict_min_dur=False,
            eth=50,
        )
        with patch("auditok.workers.os.system") as patched_os_system:
            tokenizer.start_all()
            tokenizer.join()
            tokenizer._observers[0].join()
        with open(file) as fp:
            log_lines = [
                line for line in fp.readlines() if line.startswith("[COMMAND]")
            ]

    expected_save_calls = [call(command_format) for _ in expected_detections]
    assert patched_os_system.mock_calls == expected_save_calls
    assert len(tokenizer.detections) == len(expected_detections)
    log_fmt = "[COMMAND]: Detection {id} command '{command}'"
    for i, (det, exp, log_line) in enumerate(
        zip(  # noqa: B905
            tokenizer.detections,
            expected_detections,
            log_lines,
        ),
        1,
    ):
        start, end = exp
        exp_log_line = log_fmt.format(id=i, command=command_format)
        assert pytest.approx(det.start) == start
        assert pytest.approx(det.end) == end
        assert log_line[28:].strip() == exp_log_line


def test_PrintWorker(audio_data_source, expected_detections):
    observers = [
        PrintWorker(print_format="[{id}] {start} {end}, dur: {duration}")
    ]
    tokenizer = TokenizerWorker(
        audio_data_source,
        observers=observers,
        min_dur=0.3,
        max_dur=2,
        max_silence=0.2,
        strict_min_dur=False,
        eth=50,
    )
    with patch("builtins.print") as patched_print:
        tokenizer.start_all()
        tokenizer.join()
        tokenizer._observers[0].join()

    expected_print_calls = [
        call(
            "[{}] {:.3f} {:.3f}, dur: {:.3f}".format(
                i, exp[0], exp[1], exp[1] - exp[0]
            )
        )
        for i, exp in enumerate(expected_detections, 1)
    ]
    assert patched_print.mock_calls == expected_print_calls
    assert len(tokenizer.detections) == len(expected_detections)
    for det, exp in zip(
        tokenizer.detections,
        expected_detections,
    ):
        start, end = exp
        assert pytest.approx(det.start) == start
        assert pytest.approx(det.end) == end


def test_StreamSaverWorker_wav(audio_data_source):
    with TemporaryDirectory() as tmpdir:
        expected_filename = os.path.join(tmpdir, "output.wav")
        saver = StreamSaverWorker(audio_data_source, expected_filename)
        saver.start()

        tokenizer = TokenizerWorker(saver)
        tokenizer.start_all()
        tokenizer.join()
        saver.join()

        output_filename = saver.export_audio()
        region = AudioRegion.load(
            "tests/data/test_split_10HZ_mono.raw", sr=10, sw=2, ch=1
        )

        expected_region = AudioRegion.load(output_filename)
        assert output_filename == expected_filename
        assert region == expected_region
        assert saver.data == bytes(expected_region)


@pytest.mark.parametrize(
    "export_format",
    [
        "raw",  # raw
        "wav",  # wav
    ],
    ids=[
        "raw",
        "raw",
    ],
)
def test_StreamSaverWorker(audio_data_source, export_format):
    with TemporaryDirectory() as tmpdir:
        expected_filename = os.path.join(tmpdir, f"output.{export_format}")
        saver = StreamSaverWorker(
            audio_data_source, expected_filename, export_format=export_format
        )
        saver.start()
        tokenizer = TokenizerWorker(saver)
        tokenizer.start_all()
        tokenizer.join()
        saver.join()
        output_filename = saver.export_audio()
        region = AudioRegion.load(
            "tests/data/test_split_10HZ_mono.raw", sr=10, sw=2, ch=1
        )
        expected_region = AudioRegion.load(
            output_filename, sr=10, sw=2, ch=1, audio_format=export_format
        )
        assert output_filename == expected_filename
        assert region == expected_region
        assert saver.data == bytes(expected_region)


def test_StreamSaverWorker_encode_audio(audio_data_source):
    with TemporaryDirectory() as tmpdir:
        with patch("auditok.workers._run_subprocess") as patch_rsp:
            patch_rsp.return_value = (1, None, None)
            expected_filename = os.path.join(tmpdir, "output.ogg")
            tmp_expected_filename = expected_filename + ".wav"
            saver = StreamSaverWorker(audio_data_source, expected_filename)
            saver.start()
            tokenizer = TokenizerWorker(saver)
            tokenizer.start_all()
            tokenizer.join()
            saver.join()

            with pytest.raises(auditok.workers.AudioEncodingError) as ae_error:
                saver._encode_export_audio()

        warn_msg = "Couldn't save audio data in the desired format "
        warn_msg += "'ogg'.\nEither none of 'ffmpeg', 'avconv' or 'sox' "
        warn_msg += "is installed or this format is not recognized.\n"
        warn_msg += "Audio file was saved as '{}'"
        assert warn_msg.format(tmp_expected_filename) == str(ae_error.value)
        ffmpef_avconv = [
            "-y",
            "-f",
            "wav",
            "-i",
            tmp_expected_filename,
            "-f",
            "ogg",
            expected_filename,
        ]
        expected_calls = [
            call(["ffmpeg"] + ffmpef_avconv),
            call(["avconv"] + ffmpef_avconv),
            call(
                [
                    "sox",
                    "-t",
                    "wav",
                    tmp_expected_filename,
                    expected_filename,
                ]
            ),
        ]
        assert patch_rsp.mock_calls == expected_calls
        region = AudioRegion.load(
            "tests/data/test_split_10HZ_mono.raw", sr=10, sw=2, ch=1
        )
        assert not saver._exported
        assert saver.data == bytes(region)


@pytest.mark.parametrize(
    "export_format",
    [
        "raw",  # raw
        "wav",  # wav
    ],
    ids=[
        "raw",
        "raw",
    ],
)
def test_AudioEventsJoinerWorker(audio_data_source, export_format):
    with TemporaryDirectory() as tmpdir:
        expected_filename = os.path.join(tmpdir, f"output.{export_format}")
        joiner = AudioEventsJoinerWorker(
            silence_duration=1.0,
            filename=expected_filename,
            export_format=export_format,
            sampling_rate=audio_data_source.sampling_rate,
            sample_width=audio_data_source.sample_width,
            channels=audio_data_source.channels,
        )

        tokenizer = TokenizerWorker(audio_data_source, observers=[joiner])
        tokenizer.start_all()
        tokenizer.join()
        joiner.join()

        output_filename = joiner.export_audio()
        expected_region = split_and_join_with_silence(
            "tests/data/test_split_10HZ_mono.raw",
            silence_duration=1.0,
            sr=10,
            sw=2,
            ch=1,
            aw=0.1,
        )
        assert output_filename == expected_filename
        assert joiner.data == bytes(expected_region)



from pathlib import Path

import pytest

from auditok import AudioRegion, fix_pauses, make_silence, split
from auditok.io import get_audio_source
from auditok.util import AudioReader

MONO_RAW = "tests/data/test_split_10HZ_mono.raw"
STEREO_RAW = "tests/data/test_split_10HZ_stereo.raw"
SR, SW = 10, 2

SPLIT_KWARGS_10HZ = {
    "analysis_window": 0.1,
    "eth": 50,
}


def _load_raw(filename):
    with open(filename, "rb") as fp:
        return fp.read()


def _join_regions_with_silence(regions, silence_duration, sr, sw, ch):
    """Build expected output by manually joining regions with silence."""
    silence = make_silence(silence_duration, sr, sw, ch)
    return silence.join(regions)


# ── TypeError for fixed parameters ────────────────────────────────────


class TestFixPausesRejectsFixedParams:
    """max_dur and strict_min_dur must not be accepted."""

    def test_rejects_max_dur(self):
        with pytest.raises(TypeError, match="max_dur"):
            fix_pauses(MONO_RAW, 0.5, sr=SR, sw=SW, ch=1, max_dur=10)

    def test_rejects_strict_min_dur(self):
        with pytest.raises(TypeError, match="strict_min_dur"):
            fix_pauses(MONO_RAW, 0.5, sr=SR, sw=SW, ch=1, strict_min_dur=True)

    def test_rejects_both(self):
        with pytest.raises(TypeError, match="max_dur.*strict_min_dur"):
            fix_pauses(
                MONO_RAW,
                0.5,
                sr=SR,
                sw=SW,
                ch=1,
                max_dur=10,
                strict_min_dur=True,
            )


# ── Basic behavior with raw audio data ───────────────────────────────


class TestFixPausesBasic:
    """Core behavior: joins all detected events with fixed silence."""

    def test_basic_mono(self):
        """Result should match split(max_dur=None) joined with silence."""
        data = _load_raw(MONO_RAW)
        silence_dur = 0.5
        regions = list(
            split(
                data,
                sr=SR,
                sw=SW,
                ch=1,
                min_dur=0.2,
                max_dur=None,
                max_silence=0.3,
                **SPLIT_KWARGS_10HZ,
            )
        )
        expected = _join_regions_with_silence(regions, silence_dur, SR, SW, 1)

        result = fix_pauses(
            data,
            silence_dur,
            sr=SR,
            sw=SW,
            ch=1,
            min_dur=0.2,
            max_silence=0.3,
            **SPLIT_KWARGS_10HZ,
        )
        assert bytes(result) == bytes(expected)

    def test_basic_stereo(self):
        data = _load_raw(STEREO_RAW)
        silence_dur = 0.3
        regions = list(
            split(
                data,
                sr=SR,
                sw=SW,
                ch=2,
                min_dur=0.2,
                max_dur=None,
                max_silence=0.3,
                **SPLIT_KWARGS_10HZ,
            )
        )
        expected = _join_regions_with_silence(regions, silence_dur, SR, SW, 2)

        result = fix_pauses(
            data,
            silence_dur,
            sr=SR,
            sw=SW,
            ch=2,
            min_dur=0.2,
            max_silence=0.3,
            **SPLIT_KWARGS_10HZ,
        )
        assert bytes(result) == bytes(expected)

    def test_zero_silence_duration(self):
        """With silence_duration=0, events should be concatenated."""
        data = _load_raw(MONO_RAW)
        regions = list(
            split(
                data,
                sr=SR,
                sw=SW,
                ch=1,
                min_dur=0.2,
                max_dur=None,
                max_silence=0.3,
                **SPLIT_KWARGS_10HZ,
            )
        )
        expected_data = b"".join(bytes(r) for r in regions)

        result = fix_pauses(
            data,
            0,
            sr=SR,
            sw=SW,
            ch=1,
            min_dur=0.2,
            max_silence=0.3,
            **SPLIT_KWARGS_10HZ,
        )
        assert bytes(result) == expected_data

    def test_preserves_audio_params(self):
        result = fix_pauses(
            MONO_RAW,
            0.5,
            sr=SR,
            sw=SW,
            ch=1,
            min_dur=0.2,
            max_silence=0.3,
            **SPLIT_KWARGS_10HZ,
        )
        assert result.sr == SR
        assert result.sw == SW
        assert result.ch == 1

    def test_preserves_audio_params_stereo(self):
        result = fix_pauses(
            STEREO_RAW,
            0.5,
            sr=SR,
            sw=SW,
            ch=2,
            min_dur=0.2,
            max_silence=0.3,
            **SPLIT_KWARGS_10HZ,
        )
        assert result.sr == SR
        assert result.sw == SW
        assert result.ch == 2


# ── Events are never truncated (max_dur=None internally) ─────────────


class TestFixPausesNeverTruncates:
    """Unlike split(max_dur=5), fix_pauses never splits long events."""

    def test_long_event_not_split(self):
        """A long event that split(max_dur=5) would truncate stays intact."""
        data = _load_raw(MONO_RAW)
        # With max_dur=5 and max_silence=0.2, split gives 3 regions:
        # [(2,16), (17,31), (34,76)]
        regions_with_limit = list(
            split(
                data,
                sr=SR,
                sw=SW,
                ch=1,
                min_dur=0.2,
                max_dur=5,
                max_silence=0.2,
                **SPLIT_KWARGS_10HZ,
            )
        )
        # With max_dur=None and max_silence=0.2, split gives 3 regions too
        # but with different boundaries if max_dur was the limiting factor.
        # With max_silence=10, everything becomes one event.
        regions_no_limit = list(
            split(
                data,
                sr=SR,
                sw=SW,
                ch=1,
                min_dur=0.2,
                max_dur=None,
                max_silence=10,
                **SPLIT_KWARGS_10HZ,
            )
        )
        # max_silence=10 means everything is one event
        assert len(regions_no_limit) == 1
        assert len(regions_no_limit) < len(regions_with_limit)

        # fix_pauses with max_silence=10 should produce a single event
        result = fix_pauses(
            data,
            0.5,
            sr=SR,
            sw=SW,
            ch=1,
            min_dur=0.2,
            max_silence=10,
            **SPLIT_KWARGS_10HZ,
        )
        # Single event joined with nothing = just the event itself
        assert bytes(result) == bytes(regions_no_limit[0])

    def test_consistent_with_split_max_dur_none(self):
        """fix_pauses should give the same result as
        split_and_join_with_silence when max_dur=None."""
        from auditok import split_and_join_with_silence

        data = _load_raw(MONO_RAW)
        silence_dur = 0.5
        expected = split_and_join_with_silence(
            data,
            silence_dur,
            sr=SR,
            sw=SW,
            ch=1,
            min_dur=0.2,
            max_dur=None,
            max_silence=0.3,
            strict_min_dur=False,
            **SPLIT_KWARGS_10HZ,
        )
        result = fix_pauses(
            data,
            silence_dur,
            sr=SR,
            sw=SW,
            ch=1,
            min_dur=0.2,
            max_silence=0.3,
            **SPLIT_KWARGS_10HZ,
        )
        assert bytes(result) == bytes(expected)


# ── min_dur filters short events ──────────────────────────────────────


class TestFixPausesMinDur:
    """Short events below min_dur are discarded."""

    def test_short_events_removed(self):
        """With a high min_dur, short events should be dropped."""
        data = _load_raw(MONO_RAW)
        # With min_dur=0.2, max_silence=0.2, max_dur=None we get events.
        # With min_dur=3.0, only the long event (34,76) survives.
        regions_long_min = list(
            split(
                data,
                sr=SR,
                sw=SW,
                ch=1,
                min_dur=3.0,
                max_dur=None,
                max_silence=0.2,
                **SPLIT_KWARGS_10HZ,
            )
        )
        result = fix_pauses(
            data,
            0.5,
            sr=SR,
            sw=SW,
            ch=1,
            min_dur=3.0,
            max_silence=0.2,
            **SPLIT_KWARGS_10HZ,
        )
        expected = _join_regions_with_silence(regions_long_min, 0.5, SR, SW, 1)
        assert bytes(result) == bytes(expected)
        # With min_dur=0.2 we get more events; with min_dur=3.0 fewer.
        regions_short_min = list(
            split(
                data,
                sr=SR,
                sw=SW,
                ch=1,
                min_dur=0.2,
                max_dur=None,
                max_silence=0.2,
                **SPLIT_KWARGS_10HZ,
            )
        )
        assert len(regions_long_min) < len(regions_short_min)

    def test_all_events_too_short_returns_empty(self):
        """If every event is below min_dur, return an empty AudioRegion."""
        data = _load_raw(MONO_RAW)
        result = fix_pauses(
            data,
            0.5,
            sr=SR,
            sw=SW,
            ch=1,
            min_dur=100.0,
            max_silence=0.3,
            **SPLIT_KWARGS_10HZ,
        )
        assert isinstance(result, AudioRegion)
        assert not result
        assert result.duration == 0.0
        assert result.sr == SR
        assert result.sw == SW
        assert result.ch == 1


# ── No activity returns empty AudioRegion ─────────────────────────────


class TestFixPausesNoActivity:

    def test_no_activity_returns_empty(self):
        """High energy threshold means no detections."""
        data = _load_raw(MONO_RAW)
        result = fix_pauses(
            data,
            0.5,
            sr=SR,
            sw=SW,
            ch=1,
            min_dur=0.2,
            max_silence=0.3,
            analysis_window=0.1,
            eth=60,
        )
        assert isinstance(result, AudioRegion)
        assert not result
        assert result.duration == 0.0

    def test_pure_silence_returns_empty(self):
        """Input that is all zeros should yield no events."""
        silence_data = b"\0" * SR * SW * 1 * 10  # 10 seconds of silence
        result = fix_pauses(
            silence_data,
            0.5,
            sr=SR,
            sw=SW,
            ch=1,
            min_dur=0.2,
            max_silence=0.3,
            **SPLIT_KWARGS_10HZ,
        )
        assert isinstance(result, AudioRegion)
        assert not result

    def test_empty_result_is_falsy(self):
        """Empty AudioRegion should evaluate to False like an empty list."""
        data = _load_raw(MONO_RAW)
        result = fix_pauses(
            data,
            0.5,
            sr=SR,
            sw=SW,
            ch=1,
            min_dur=100.0,
            max_silence=0.3,
            **SPLIT_KWARGS_10HZ,
        )
        assert not result
        # Non-empty result should be truthy
        result_nonempty = fix_pauses(
            data,
            0.5,
            sr=SR,
            sw=SW,
            ch=1,
            min_dur=0.2,
            max_silence=0.3,
            **SPLIT_KWARGS_10HZ,
        )
        assert result_nonempty


# ── max_leading_silence and max_trailing_silence ──────────────────────


class TestFixPausesSilenceParams:

    def test_max_trailing_silence_zero(self):
        """Trailing silence should be stripped from each event."""
        data = _load_raw(MONO_RAW)
        regions = list(
            split(
                data,
                sr=SR,
                sw=SW,
                ch=1,
                min_dur=0.2,
                max_dur=None,
                max_silence=0.3,
                max_trailing_silence=0,
                **SPLIT_KWARGS_10HZ,
            )
        )
        expected = _join_regions_with_silence(regions, 0.5, SR, SW, 1)

        result = fix_pauses(
            data,
            0.5,
            sr=SR,
            sw=SW,
            ch=1,
            min_dur=0.2,
            max_silence=0.3,
            max_trailing_silence=0,
            **SPLIT_KWARGS_10HZ,
        )
        assert bytes(result) == bytes(expected)

    def test_max_trailing_silence_changes_output(self):
        """Trimming trailing silence should produce shorter output."""
        data = _load_raw(MONO_RAW)
        result_with = fix_pauses(
            data,
            0.5,
            sr=SR,
            sw=SW,
            ch=1,
            min_dur=0.2,
            max_silence=0.3,
            **SPLIT_KWARGS_10HZ,
        )
        result_trimmed = fix_pauses(
            data,
            0.5,
            sr=SR,
            sw=SW,
            ch=1,
            min_dur=0.2,
            max_silence=0.3,
            max_trailing_silence=0,
            **SPLIT_KWARGS_10HZ,
        )
        # Trimming trailing silence from events makes the total shorter
        assert result_trimmed.duration <= result_with.duration

    def test_max_leading_silence(self):
        """Leading silence should be prepended to each event."""
        data = _load_raw(MONO_RAW)
        result_no_lead = fix_pauses(
            data,
            0.5,
            sr=SR,
            sw=SW,
            ch=1,
            min_dur=0.2,
            max_silence=0.3,
            max_leading_silence=0,
            **SPLIT_KWARGS_10HZ,
        )
        result_with_lead = fix_pauses(
            data,
            0.5,
            sr=SR,
            sw=SW,
            ch=1,
            min_dur=0.2,
            max_silence=0.3,
            max_leading_silence=0.2,
            **SPLIT_KWARGS_10HZ,
        )
        # Adding leading silence to events makes the total longer
        assert result_with_lead.duration >= result_no_lead.duration


# ── Silence duration variations ───────────────────────────────────────


class TestFixPausesSilenceDuration:

    @pytest.mark.parametrize(
        "silence_dur",
        [0, 0.1, 0.5, 1.0, 2.0],
        ids=["zero", "short", "medium", "one_sec", "long"],
    )
    def test_various_silence_durations(self, silence_dur):
        """Output should match manual join for any silence duration."""
        data = _load_raw(MONO_RAW)
        regions = list(
            split(
                data,
                sr=SR,
                sw=SW,
                ch=1,
                min_dur=0.2,
                max_dur=None,
                max_silence=0.3,
                **SPLIT_KWARGS_10HZ,
            )
        )
        expected = _join_regions_with_silence(regions, silence_dur, SR, SW, 1)

        result = fix_pauses(
            data,
            silence_dur,
            sr=SR,
            sw=SW,
            ch=1,
            min_dur=0.2,
            max_silence=0.3,
            **SPLIT_KWARGS_10HZ,
        )
        assert bytes(result) == bytes(expected)


# ── Input types ───────────────────────────────────────────────────────


class TestFixPausesInputTypes:
    """fix_pauses should work with all supported input types."""

    _COMMON_KW = {
        "min_dur": 0.2,
        "max_silence": 0.3,
        **SPLIT_KWARGS_10HZ,
    }

    def _reference(self):
        data = _load_raw(MONO_RAW)
        return bytes(
            fix_pauses(data, 0.5, sr=SR, sw=SW, ch=1, **self._COMMON_KW)
        )

    def test_filename(self):
        result = fix_pauses(
            MONO_RAW, 0.5, sr=SR, sw=SW, ch=1, **self._COMMON_KW
        )
        assert bytes(result) == self._reference()

    def test_path(self):
        result = fix_pauses(
            Path(MONO_RAW), 0.5, sr=SR, sw=SW, ch=1, **self._COMMON_KW
        )
        assert bytes(result) == self._reference()

    def test_bytes(self):
        data = _load_raw(MONO_RAW)
        result = fix_pauses(data, 0.5, sr=SR, sw=SW, ch=1, **self._COMMON_KW)
        assert bytes(result) == self._reference()

    def test_audio_region(self):
        data = _load_raw(MONO_RAW)
        region = AudioRegion(data, SR, SW, 1)
        result = fix_pauses(region, 0.5, **self._COMMON_KW)
        assert bytes(result) == self._reference()

    def test_audio_source(self):
        source = get_audio_source(MONO_RAW, sr=SR, sw=SW, ch=1)
        result = fix_pauses(source, 0.5, **self._COMMON_KW)
        assert bytes(result) == self._reference()

    def test_audio_reader(self):
        reader = AudioReader(
            MONO_RAW, sr=SR, sw=SW, ch=1, block_dur=0.1, record=True
        )
        result = fix_pauses(reader, 0.5, **self._COMMON_KW)
        assert bytes(result) == self._reference()



"""
:author:

Amine SEHILI <amine.sehili@gmail.com>
2015-2026

:License:

This package is published under the MIT license.
"""

from .audio import *
from .core import *
from .exceptions import *
from .io import *
from .util import *

__version__ = "0.4.0"



#!/usr/bin/env python
# encoding: utf-8
"""`auditok` -- An Audio Activity Detection Tool

`auditok` is a program designed for audio or acoustic activity detection.
It supports reading audio data from various sources, including audio files,
microphones, and standard input.

@author:     Mohamed El Amine SEHILI
@copyright:  2015-2026 Mohamed El Amine SEHILI
@license:    MIT
@contact:    amine.sehili@gmail.com
@deffield    updated: 01 Apr 2026
"""

import os
import sys
import tempfile
import threading
import time
from argparse import ArgumentParser

from auditok import AudioRegion, __version__

from .audio import fix_pauses, trim
from .cmdline_util import initialize_workers, make_kwargs, make_logger
from .exceptions import ArgumentError, EndOfProcessing

__all__ = []  # type: ignore[var-annotated]
__date__ = "2015-11-23"
__updated__ = "2026-04-01"

_SUBCOMMANDS = {"split", "trim", "fix-pauses"}

# ANSI escape helpers
_BOLD = "\033[1m"
_DIM = "\033[2m"
_ITALIC = "\033[3m"
_RED = "\033[91m"
_RESET = "\033[0m"
_CLEAR_LINE = "\033[2K\r"


def _recording_loop(quiet=False):
    """Block while workers are running, showing a recording indicator."""
    start = time.monotonic()
    dot_visible = True
    try:
        while True:
            if not quiet:
                elapsed = time.monotonic() - start
                mins, secs = divmod(int(elapsed), 60)
                dot = f"{_RED}\u25cf{_RESET}" if dot_visible else " "
                sys.stderr.write(
                    f"{_CLEAR_LINE}{dot} {_DIM}{_ITALIC}Recording "
                    f"{mins:02d}:{secs:02d}  "
                    f"(Ctrl+C to stop){_RESET}"
                )
                sys.stderr.flush()
                dot_visible = not dot_visible
            time.sleep(1)
            if len(threading.enumerate()) == 1:
                raise EndOfProcessing
    except (KeyboardInterrupt, EndOfProcessing):
        if not quiet:
            sys.stderr.write(f"{_CLEAR_LINE}")
            sys.stderr.flush()


# ── Shared argument helpers ──────────────────────────────────────────


def _add_input_source_args(group):
    """Add input source configuration arguments."""
    group.add_argument(
        "-I",
        "--input-device-index",
        dest="input_device_index",
        help="Audio device index [Default: %(default)s]. "
        "Optional and only effective when reading from a microphone.",
        type=int,
        default=None,
        metavar="INT",
    )
    group.add_argument(
        "-F",
        "--audio-frame-per-buffer",
        dest="frame_per_buffer",
        help="Audio frame per buffer [Default: %(default)s]. "
        "Optional and only effective when reading from a microphone.",
        type=int,
        default=1024,
        metavar="INT",
    )
    group.add_argument(
        "-f",
        "--input-format",
        dest="input_format",
        type=str,
        default=None,
        help="Specify the input audio file format. If not provided, the "
        "format is inferred from the file extension. If the output file "
        "name lacks an extension, the format is guessed from the file "
        "header (requires ffmpeg). If neither condition is met, an error "
        "is raised.",
        metavar="STRING",
    )
    group.add_argument(
        "-M",
        "--max-read",
        dest="max_read",
        type=float,
        default=None,
        help="Maximum data (in seconds) to read from a microphone or a file"
        " [Default: read until the end of the file or stream].",
        metavar="FLOAT",
    )
    group.add_argument(
        "-L",
        "--large-file",
        dest="large_file",
        action="store_true",
        default=False,
        help="Whether the input file should be treated as a large file. "
        "If True, data will be read from file on demand, otherwise all "
        "audio data is loaded into memory before tokenization.",
    )


def _add_output_format_arg(group):
    """Add -T/--output-format argument."""
    group.add_argument(
        "-T",
        "--output-format",
        dest="output_format",
        type=str,
        default=None,
        help="Specify the audio format for saving detections and/or the "
        "main stream. If not provided, the format will be (1) inferred from"
        " the file extension or (2) default to raw format.",
        metavar="STRING",
    )


def _add_channel_arg(group):
    """Add -u/--use-channel argument."""
    group.add_argument(
        "-u",
        "--use-channel",
        dest="use_channel",
        type=str,
        default=None,
        help="Specify the audio channel to use for tokenization when the "
        "input stream is multi-channel (0 refers to the first channel). By "
        "default, this is set to None, meaning all channels are used, "
        "capturing any valid audio event from any channel. Alternatively, "
        "set this to 'mix' (or 'avg'/'average') to combine all channels "
        "into a single averaged channel for tokenization. Regardless of the"
        " option chosen, saved audio events will have the same number of "
        "channels as the input stream. [Default: %(default)s, use all "
        "channels].",
        metavar="INT/STRING",
    )


def _add_detection_args(group):
    """Add common detection arguments."""
    group.add_argument(
        "-a",
        "--analysis-window",
        dest="analysis_window",
        default=0.01,
        type=float,
        help="Specify the size of the analysis window in seconds. "
        "[Default: %(default)s (10ms)].",
        metavar="FLOAT",
    )
    group.add_argument(
        "-n",
        "--min-duration",
        dest="min_duration",
        type=float,
        default=0.2,
        help="Minimum duration of a valid audio event in seconds. "
        "[Default: %(default)s].",
        metavar="FLOAT",
    )
    group.add_argument(
        "-s",
        "--max-silence",
        dest="max_silence",
        type=float,
        default=0.3,
        help="Maximum duration of consecutive silence allowed within a "
        "valid audio event in seconds. [Default: %(default)s]",
        metavar="FLOAT",
    )
    group.add_argument(
        "-l",
        "--max-leading-silence",
        dest="max_leading_silence",
        type=float,
        default=0,
        help="Maximum duration (in seconds) of silence to retain before "
        "each detected event. Preserves the natural onset of sounds "
        "(e.g., the gradual rise of speech). A value of 0.1-0.3 seconds "
        "is typically a good choice. [Default: %(default)s]",
        metavar="FLOAT",
    )
    group.add_argument(
        "-g",
        "--max-trailing-silence",
        dest="max_trailing_silence",
        type=float,
        default=None,
        help="Maximum duration (in seconds) of trailing silence to keep "
        "at the end of each detected event. Use 0 to drop all trailing "
        "silence. When omitted, all trailing silence (up to --max-silence) "
        "is kept. [Default: %(default)s]",
        metavar="FLOAT",
    )
    group.add_argument(
        "-e",
        "--energy-threshold",
        dest="energy_threshold",
        type=float,
        default=50,
        help="Set the log energy threshold for detection. "
        "[Default: %(default)s]",
        metavar="FLOAT",
    )


def _add_audio_params(group):
    """Add audio parameter arguments."""
    group.add_argument(
        "-r",
        "--rate",
        dest="sampling_rate",
        type=int,
        default=16000,
        help="Sampling rate of audio data [Default: %(default)s].",
        metavar="INT",
    )
    group.add_argument(
        "-c",
        "--channels",
        dest="channels",
        type=int,
        default=1,
        help="Number of channels of audio data [Default: %(default)s].",
        metavar="INT",
    )
    group.add_argument(
        "-w",
        "--width",
        dest="sample_width",
        type=int,
        default=2,
        help="Number of bytes per audio sample [Default: %(default)s].",
        metavar="INT",
    )


def _add_debug_args(parser):
    """Add debug arguments."""
    parser.add_argument(
        "-D",
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Debug mode: output processing operations to STDOUT.",
    )
    parser.add_argument(
        "--debug-file",
        dest="debug_file",
        type=str,
        default=None,
        help="Save processing operations to the specified file.",
        metavar="FILE",
    )


# ── Subparser setup ──────────────────────────────────────────────────


def _setup_split_parser(subparsers):
    """Set up the 'split' subcommand parser."""
    parser = subparsers.add_parser(
        "split",
        usage="auditok split [input] [options]",
        description="Detect and split audio into individual events based on "
        "energy thresholds. This is the default subcommand.",
        help="Detect and split audio into events (default subcommand).",
    )

    group = parser.add_argument_group("Input-Output options")
    group.add_argument(
        dest="input",
        help="Input audio or video file. Use '-' for stdin "
        "[Default: read from a microphone using sounddevice].",
        metavar="input",
        nargs="?",
        default=None,
    )
    _add_input_source_args(group)
    group.add_argument(
        "-O",
        "--save-stream",
        dest="save_stream",
        type=str,
        default=None,
        help="Save read audio data (from a file or a microphone) to a file."
        " If omitted, no audio data will be saved.",
        metavar="FILE",
    )
    group.add_argument(
        "-o",
        "--save-detections-as",
        dest="save_detections_as",
        type=str,
        default=None,
        help="Specify the file name format to save detected events. You can "
        "use the following placeholders to construct the output file name: "
        "{id} (sequential, starting from 1), {start}, {end}, and {duration}. "
        "Time placeholders are in seconds. "
        "Example: 'Event_{id}{start}-{end}{duration:.3f}.wav'",
        metavar="STRING",
    )
    group.add_argument(
        "-j",
        "--join-detections",
        dest="join_detections",
        type=float,
        default=None,
        help="[Deprecated: use 'auditok fix-pauses' instead.] "
        "Join (glue) detected audio events with a specified duration "
        "of silence between them. To be used in combination with the "
        "--save-stream / -O option.",
        metavar="FLOAT",
    )
    _add_output_format_arg(group)
    _add_channel_arg(group)

    group = parser.add_argument_group(
        "Tokenization options",
        "Set audio events' duration and set the threshold for detection.",
    )
    _add_detection_args(group)
    group.add_argument(
        "-m",
        "--max-duration",
        dest="max_duration",
        type=float,
        default=5,
        help="Maximum duration of a valid audio event in seconds. "
        "[Default: %(default)s].",
        metavar="FLOAT",
    )
    group.add_argument(
        "-d",
        "--drop-trailing-silence",
        dest="drop_trailing_silence",
        action="store_true",
        default=False,
        help="[Deprecated: use -g/--max-trailing-silence 0 instead.] "
        "Remove trailing silence from a detection. Ignored if "
        "-g/--max-trailing-silence is also provided. "
        "[Default: trailing silence is retained].",
    )
    group.add_argument(
        "-R",
        "--strict-min-duration",
        dest="strict_min_duration",
        action="store_true",
        default=False,
        help="Reject events shorter than --min-duration, even if adjacent "
        "to the most recent valid event that reached max-duration. "
        "[Default: retain such events].",
    )

    group = parser.add_argument_group(
        "Audio parameters",
        "Set audio parameters when reading from a headerless file "
        "(raw or stdin) or when using custom microphone settings.",
    )
    _add_audio_params(group)

    group = parser.add_argument_group(
        "Use audio events",
        "Use these options to print, play, or plot detected audio events.",
    )
    group.add_argument(
        "-C",
        "--command",
        dest="command",
        type=str,
        help="Provide a command to execute when an audio event is detected."
        " Use '{file}' as a placeholder for the temporary WAV file "
        "containing the event data (e.g., `-C 'du -h {file}'` to "
        "display the file size or `-C 'play -q {file}'` to play audio "
        "with sox).",
        metavar="STRING",
    )
    group.add_argument(
        "-E",
        "--echo",
        dest="echo",
        action="store_true",
        default=False,
        help="Immediately play back a detected audio event using "
        "sounddevice.",
    )
    group.add_argument(
        "-B",
        "--progress-bar",
        dest="progress_bar",
        action="store_true",
        default=False,
        help="Show a progress bar when playing audio.",
    )
    group.add_argument(
        "-p",
        "--plot",
        dest="plot",
        action="store_true",
        default=False,
        help="Plot and displays the audio signal along with detections "
        "(requires matplotlib).",
    )
    group.add_argument(
        "--save-image",
        dest="save_image",
        type=str,
        help="Save the plotted audio signal and detections as a picture "
        "or a PDF file (requires matplotlib).",
        metavar="FILE",
    )
    group.add_argument(
        "--printf",
        dest="printf",
        type=str,
        default="{id} {start} {end}",
        help="Prints information about each audio event on a new line "
        "using the specified format. The format can include text and "
        "placeholders: {id} (sequential, starting from 1), {start}, "
        "{end}, {duration}, and {timestamp}. The first three time "
        "placeholders are in seconds, with formatting controlled by the "
        "--time-format argument. {timestamp} represents the system date "
        "and time of the event, configurable with the --timestamp-format "
        "argument. Example: '[{id}]: {start} -> {end} -- {timestamp}'.",
        metavar="STRING",
    )
    group.add_argument(
        "--time-format",
        dest="time_format",
        type=str,
        default="%S",
        help="Specify the format for printing {start}, {end}, and "
        "{duration} placeholders with --printf. [Default: %(default)s]. "
        "Accepted formats are\n:"
        " - %%S: absolute time in seconds\n"
        " - %%I: absolute time in milliseconds\n"
        " - %%h, %%m, %%s, %%i: converts time into hours, minutes, seconds,"
        " and milliseconds (e.g., %%h:%%m:%%s.%%i) and only displays "
        "provided fields.\nNote that %%S and %%I can only be used "
        "independently.",
        metavar="STRING",
    )
    group.add_argument(
        "--timestamp-format",
        dest="timestamp_format",
        type=str,
        default="%Y/%m/%d %H:%M:%S",
        help="Specify the format used for printing {timestamp}. Should be "
        "a format accepted by the 'datetime' standard module. [Default: "
        "'%%Y/%%m/%%d %%H:%%M:%%S'].",
    )

    parser.add_argument(
        "-q",
        "--quiet",
        dest="quiet",
        action="store_true",
        default=False,
        help="Quiet mode: Do not display any information on the screen.",
    )
    _add_debug_args(parser)

    parser.set_defaults(func=_run_split)


def _setup_trim_parser(subparsers):
    """Set up the 'trim' subcommand parser."""
    parser = subparsers.add_parser(
        "trim",
        usage="auditok trim [input] -o output [options]",
        description="Remove leading and trailing silence from audio, "
        "keeping everything between the first and last detected events.",
        help="Remove leading and trailing silence from audio.",
    )

    group = parser.add_argument_group("Input-Output options")
    group.add_argument(
        dest="input",
        help="Input audio or video file. Use '-' for stdin "
        "[Default: read from a microphone using sounddevice].",
        metavar="input",
        nargs="?",
        default=None,
    )
    group.add_argument(
        "-o",
        "--output",
        dest="output",
        type=str,
        required=True,
        help="Output file path.",
        metavar="FILE",
    )
    _add_input_source_args(group)
    _add_output_format_arg(group)
    _add_channel_arg(group)

    group = parser.add_argument_group(
        "Tokenization options",
        "Set detection parameters for identifying audio events.",
    )
    _add_detection_args(group)

    group = parser.add_argument_group(
        "Audio parameters",
        "Set audio parameters when reading from a headerless file "
        "(raw or stdin) or when using custom microphone settings.",
    )
    _add_audio_params(group)

    parser.add_argument(
        "-q",
        "--quiet",
        dest="quiet",
        action="store_true",
        default=False,
        help="Quiet mode: suppress recording indicator and status messages.",
    )
    _add_debug_args(parser)

    parser.set_defaults(func=_run_trim)


def _setup_fix_pauses_parser(subparsers):
    """Set up the 'fix-pauses' subcommand parser."""
    parser = subparsers.add_parser(
        "fix-pauses",
        usage="auditok fix-pauses [input] -o output -d pause_duration [options]",
        description="Normalize pauses between detected audio events to a "
        "fixed duration, removing excess silence.",
        help="Normalize pauses between audio events to a fixed duration.",
    )

    group = parser.add_argument_group("Input-Output options")
    group.add_argument(
        dest="input",
        help="Input audio or video file. Use '-' for stdin "
        "[Default: read from a microphone using sounddevice].",
        metavar="input",
        nargs="?",
        default=None,
    )
    group.add_argument(
        "-o",
        "--output",
        dest="output",
        type=str,
        required=True,
        help="Output file path.",
        metavar="FILE",
    )
    group.add_argument(
        "-d",
        "--pause-duration",
        dest="pause_duration",
        type=float,
        required=True,
        help="Duration of silence (in seconds) to insert between detected "
        "audio events.",
        metavar="FLOAT",
    )
    _add_input_source_args(group)
    _add_output_format_arg(group)
    _add_channel_arg(group)

    group = parser.add_argument_group(
        "Tokenization options",
        "Set detection parameters for identifying audio events.",
    )
    _add_detection_args(group)

    group = parser.add_argument_group(
        "Audio parameters",
        "Set audio parameters when reading from a headerless file "
        "(raw or stdin) or when using custom microphone settings.",
    )
    _add_audio_params(group)

    parser.add_argument(
        "-q",
        "--quiet",
        dest="quiet",
        action="store_true",
        default=False,
        help="Quiet mode: suppress recording indicator and status messages.",
    )
    _add_debug_args(parser)

    parser.set_defaults(func=_run_fix_pauses)


# ── Handler functions ────────────────────────────────────────────────


def _parse_use_channel(value):
    try:
        return int(value)
    except (ValueError, TypeError):
        return value


def _build_kwargs(args):
    """Build split and audio kwargs dicts from parsed args."""
    use_channel = _parse_use_channel(args.use_channel)
    split_kw = {
        "min_dur": args.min_duration,
        "max_silence": args.max_silence,
        "max_leading_silence": args.max_leading_silence,
        "max_trailing_silence": args.max_trailing_silence,
        "energy_threshold": args.energy_threshold,
        "analysis_window": args.analysis_window,
    }
    audio_kw = {
        "sampling_rate": args.sampling_rate,
        "sample_width": args.sample_width,
        "channels": args.channels,
        "use_channel": use_channel,
        "audio_format": args.input_format,
        "large_file": args.large_file,
        "max_read": args.max_read,
        "frames_per_buffer": args.frame_per_buffer,
        "input_device_index": args.input_device_index,
    }
    return split_kw, audio_kw


def _run_split(args):
    """Execute the split subcommand (original auditok behavior)."""
    try:
        kwargs = make_kwargs(args)
    except ArgumentError as exc:
        print(exc, file=sys.stderr)
        return 1

    if args.join_detections is not None:
        print(
            "Warning: -j/--join-detections is deprecated. "
            "Use 'auditok fix-pauses' instead.",
            file=sys.stderr,
        )

    logger = make_logger(args.debug, args.debug_file)
    tokenizer_worker = None
    stream_saver = None

    try:
        stream_saver, tokenizer_worker = initialize_workers(
            logger=logger, **kwargs.split, **kwargs.io, **kwargs.miscellaneous
        )
        tokenizer_worker.start_all()

        while True:
            time.sleep(1)
            if len(threading.enumerate()) == 1:
                raise EndOfProcessing

    except (KeyboardInterrupt, EndOfProcessing):
        if tokenizer_worker is not None:
            tokenizer_worker.stop_all()

            if stream_saver is not None:
                stream_saver.join()
                try:
                    stream_saver.export_audio()
                except Exception as aee:
                    print(aee, file=sys.stderr)

            if args.plot or args.save_image is not None:
                from .plotting import plot

                reader = tokenizer_worker.reader

                reader.rewind()
                record = AudioRegion(
                    reader.data, reader.sr, reader.sw, reader.ch
                )
                detections = (
                    (det.start, det.end) for det in tokenizer_worker.detections
                )
                plot(
                    record,
                    detections=detections,
                    energy_threshold=args.energy_threshold,
                    show=True,
                    save_as=args.save_image,
                )
    return 0


def _run_trim(args):
    """Execute the trim subcommand."""
    logger = make_logger(args.debug, args.debug_file)
    split_kw, audio_kw = _build_kwargs(args)

    if args.input is not None:
        # File input: call trim() directly
        result = trim(args.input, **split_kw, **audio_kw)
        if result:
            result.save(args.output, audio_format=args.output_format)
        else:
            print("No audio activity detected.", file=sys.stderr)
        return 0

    # Mic input: worker-based approach
    from . import workers
    from .util import AudioReader

    reader = AudioReader(
        None,
        block_dur=split_kw["analysis_window"],
        sampling_rate=audio_kw["sampling_rate"],
        sample_width=audio_kw["sample_width"],
        channels=audio_kw["channels"],
        frames_per_buffer=audio_kw["frames_per_buffer"],
        input_device_index=audio_kw["input_device_index"],
        max_read=audio_kw["max_read"],
    )

    tmp_fd, tmp_path = tempfile.mkstemp(suffix=".wav")
    os.close(tmp_fd)
    stream_saver = workers.StreamSaverWorker(
        reader, filename=tmp_path, export_format="wav"
    )
    stream_saver.start()

    tokenizer_worker = workers.TokenizerWorker(
        stream_saver,
        observers=[],
        logger=logger,
        max_dur=None,
        strict_min_dur=False,
        use_channel=audio_kw["use_channel"],
        **split_kw,
    )
    tokenizer_worker.start_all()
    _recording_loop(quiet=args.quiet)
    tokenizer_worker.stop_all()

    detections = tokenizer_worker.detections
    if detections:
        first, last = detections[0], detections[-1]
        data = stream_saver.data
        full_region = AudioRegion(
            data, stream_saver.sr, stream_saver.sw, stream_saver.ch
        )
        trimmed = full_region.sec[first.start : last.end]
        trimmed.save(args.output, audio_format=args.output_format)
        if not args.quiet:
            print(
                f"{_DIM}Saved to {_RESET}{_BOLD}{args.output}{_RESET}",
                file=sys.stderr,
            )
    else:
        if not args.quiet:
            print("No audio activity detected.", file=sys.stderr)

    try:
        os.unlink(tmp_path)
    except OSError:
        pass

    return 0


def _run_fix_pauses(args):
    """Execute the fix-pauses subcommand."""
    logger = make_logger(args.debug, args.debug_file)
    split_kw, audio_kw = _build_kwargs(args)

    if args.input is not None:
        # File input: call fix_pauses() directly
        result = fix_pauses(
            args.input, args.pause_duration, **split_kw, **audio_kw
        )
        if result:
            result.save(args.output, audio_format=args.output_format)
        else:
            print("No audio activity detected.", file=sys.stderr)
        return 0

    # Mic input: worker-based approach with AudioEventsJoinerWorker
    from . import workers
    from .util import AudioReader

    reader = AudioReader(
        None,
        block_dur=split_kw["analysis_window"],
        sampling_rate=audio_kw["sampling_rate"],
        sample_width=audio_kw["sample_width"],
        channels=audio_kw["channels"],
        frames_per_buffer=audio_kw["frames_per_buffer"],
        input_device_index=audio_kw["input_device_index"],
        max_read=audio_kw["max_read"],
    )

    joiner = workers.AudioEventsJoinerWorker(
        silence_duration=args.pause_duration,
        filename=args.output,
        export_format=args.output_format,
        sampling_rate=reader.sampling_rate,
        sample_width=reader.sample_width,
        channels=reader.channels,
    )

    tokenizer_worker = workers.TokenizerWorker(
        reader,
        observers=[joiner],
        logger=logger,
        max_dur=None,
        strict_min_dur=False,
        use_channel=audio_kw["use_channel"],
        **split_kw,
    )
    tokenizer_worker.start_all()
    _recording_loop(quiet=args.quiet)
    tokenizer_worker.stop_all()

    try:
        joiner.export_audio()
        if not args.quiet:
            print(
                f"{_DIM}Saved to {_RESET}{_BOLD}{args.output}{_RESET}",
                file=sys.stderr,
            )
    except Exception as exc:
        if not args.quiet:
            print(exc, file=sys.stderr)

    return 0


# ── Main entry point ─────────────────────────────────────────────────


def main(argv=None):
    program_name = os.path.basename(sys.argv[0])
    if argv is None:
        argv = sys.argv[1:]

    # Handle --version/-v before subcommand dispatch
    if argv and argv[0] in ("--version", "-v"):
        print(__version__)
        return 0

    # Bare "auditok" → implicit "split" (start recording from mic)
    # "auditok file.wav ..." → implicit "split file.wav ..."
    if not argv:
        argv = ["split"]
    elif argv[0] in ("--help", "-h"):
        pass  # Let main parser show subcommand listing
    elif argv[0] not in _SUBCOMMANDS:
        argv = ["split"] + argv

    parser = ArgumentParser(
        prog=program_name,
        description="auditok, an audio tokenization tool.",
        epilog="Run '%(prog)s <command> -h' for help on a specific command. "
        "Running '%(prog)s' without a command is equivalent to "
        "'%(prog)s split'.",
    )
    subparsers = parser.add_subparsers(dest="subcommand")
    _setup_split_parser(subparsers)
    _setup_trim_parser(subparsers)
    _setup_fix_pauses_parser(subparsers)

    args = parser.parse_args(argv)

    if not hasattr(args, "func"):
        parser.print_help()
        return 0

    return args.func(args)


if __name__ == "__main__":
    sys.exit(main(None))



from pathlib import Path

import pytest

from auditok import AudioRegion, split, trim
from auditok.io import get_audio_source
from auditok.util import AudioReader

SPLIT_KWARGS_10HZ = {
    "min_dur": 0.2,
    "max_dur": 5,
    "max_silence": 0.2,
    "strict_min_dur": False,
    "analysis_window": 0.1,
    "eth": 50,
}

# trim() no longer accepts max_dur or strict_min_dur
TRIM_KWARGS_10HZ = {
    "min_dur": 0.2,
    "max_silence": 0.2,
    "analysis_window": 0.1,
    "eth": 50,
}

MONO_RAW = "tests/data/test_split_10HZ_mono.raw"
STEREO_RAW = "tests/data/test_split_10HZ_stereo.raw"
SR, SW = 10, 2


def _load_raw(filename):
    with open(filename, "rb") as fp:
        return fp.read()


def _expected_trimmed(data, channels, first_onset, last_offset):
    """Return the expected bytes for a trim between first_onset and
    last_offset (in samples)."""
    sample_size = SW * channels
    return data[first_onset * sample_size : last_offset * sample_size]


# ── AudioRegion.trim() ──────────────────────────────────────────────


class TestAudioRegionTrim:
    """Tests for AudioRegion.trim() method."""

    def test_trim_basic_mono(self):
        """Trim should remove leading and trailing silence."""
        data = _load_raw(MONO_RAW)
        region = AudioRegion(data, SR, SW, 1)
        trimmed = region.trim(**TRIM_KWARGS_10HZ)
        expected = _expected_trimmed(data, 1, 2, 76)
        assert bytes(trimmed) == expected

    def test_trim_basic_stereo(self):
        """Trim should work with stereo data."""
        data = _load_raw(STEREO_RAW)
        region = AudioRegion(data, SR, SW, 2)
        trimmed = region.trim(**TRIM_KWARGS_10HZ)
        expected = _expected_trimmed(data, 2, 2, 76)
        assert bytes(trimmed) == expected

    def test_trim_preserves_audio_params(self):
        """Trimmed region should keep sr, sw, ch from the original."""
        data = _load_raw(STEREO_RAW)
        region = AudioRegion(data, SR, SW, 2)
        trimmed = region.trim(**TRIM_KWARGS_10HZ)
        assert trimmed.sr == SR
        assert trimmed.sw == SW
        assert trimmed.ch == 2

    def test_trim_no_activity_returns_empty_region(self):
        """When no activity is detected, return an empty AudioRegion."""
        data = _load_raw(MONO_RAW)
        region = AudioRegion(data, SR, SW, 1)
        trimmed = region.trim(
            **{**TRIM_KWARGS_10HZ, "eth": 60},
        )
        assert isinstance(trimmed, AudioRegion)
        assert trimmed.duration == 0.0
        assert trimmed.sr == SR
        assert trimmed.sw == SW
        assert trimmed.ch == 1

    def test_trim_empty_region_is_joinable(self):
        """Empty region from trim should be usable with + and join."""
        data = _load_raw(MONO_RAW)
        region = AudioRegion(data, SR, SW, 1)
        empty = region.trim(**{**TRIM_KWARGS_10HZ, "eth": 60})
        joined = empty + region
        assert joined.duration == region.duration

    def test_trim_consistent_with_split(self):
        """trim() result should span from first split start to last split end."""
        data = _load_raw(MONO_RAW)
        region = AudioRegion(data, SR, SW, 1)
        regions = list(
            region.split(max_dur=None, strict_min_dur=False, **TRIM_KWARGS_10HZ)
        )
        trimmed = region.trim(**TRIM_KWARGS_10HZ)
        first, last = regions[0], regions[-1]
        expected = region.sec[first.start : last.end]
        assert bytes(trimmed) == bytes(expected)

    def test_trim_preserves_internal_silence(self):
        """Silence between detections must be preserved."""
        data = _load_raw(MONO_RAW)
        region = AudioRegion(data, SR, SW, 1)
        trimmed = region.trim(**TRIM_KWARGS_10HZ)
        regions = list(
            region.split(max_dur=None, strict_min_dur=False, **TRIM_KWARGS_10HZ)
        )
        detection_dur = sum(r.duration for r in regions)
        assert trimmed.duration > detection_dur

    @pytest.mark.parametrize(
        "min_dur, max_silence, max_trailing_silence, "
        "kwargs, expected_onset, expected_offset",
        [
            (0.2, 0.2, None, {"eth": 50}, 2, 76),
            (3, 0.2, None, {"eth": 50}, 34, 76),
            (0.2, 10, None, {"eth": 50}, 2, 76),
            (0.2, 0.2, 0, {"eth": 50}, 2, 76),
            (0.2, 0.2, None, {"energy_threshold": 40}, 0, 76),
        ],
        ids=[
            "default",
            "long_min_dur",
            "long_max_silence",
            "drop_trailing_silence",
            "low_energy_threshold",
        ],
    )
    def test_trim_with_various_params(
        self,
        min_dur,
        max_silence,
        max_trailing_silence,
        kwargs,
        expected_onset,
        expected_offset,
    ):
        """Trim boundaries should match first/last detection boundaries."""
        data = _load_raw(MONO_RAW)
        region = AudioRegion(data, SR, SW, 1)
        trim_kwargs = dict(
            min_dur=min_dur,
            max_silence=max_silence,
            analysis_window=0.1,
            **kwargs,
        )
        if max_trailing_silence is not None:
            trim_kwargs["max_trailing_silence"] = max_trailing_silence
        trimmed = region.trim(**trim_kwargs)
        expected = _expected_trimmed(data, 1, expected_onset, expected_offset)
        assert bytes(trimmed) == expected


# ── Module-level trim() ─────────────────────────────────────────────


class TestModuleTrim:
    """Tests for the module-level trim() function."""

    def test_trim_with_filename(self):
        data = _load_raw(MONO_RAW)
        trimmed = trim(MONO_RAW, sr=SR, sw=SW, ch=1, **TRIM_KWARGS_10HZ)
        expected = _expected_trimmed(data, 1, 2, 76)
        assert bytes(trimmed) == expected

    def test_trim_with_path(self):
        data = _load_raw(MONO_RAW)
        trimmed = trim(Path(MONO_RAW), sr=SR, sw=SW, ch=1, **TRIM_KWARGS_10HZ)
        expected = _expected_trimmed(data, 1, 2, 76)
        assert bytes(trimmed) == expected

    def test_trim_with_bytes(self):
        data = _load_raw(MONO_RAW)
        trimmed = trim(data, sr=SR, sw=SW, ch=1, **TRIM_KWARGS_10HZ)
        expected = _expected_trimmed(data, 1, 2, 76)
        assert bytes(trimmed) == expected

    def test_trim_with_audio_region(self):
        data = _load_raw(MONO_RAW)
        region = AudioRegion(data, SR, SW, 1)
        trimmed = trim(region, **TRIM_KWARGS_10HZ)
        expected = _expected_trimmed(data, 1, 2, 76)
        assert bytes(trimmed) == expected

    def test_trim_with_audio_source(self):
        data = _load_raw(MONO_RAW)
        source = get_audio_source(MONO_RAW, sr=SR, sw=SW, ch=1)
        trimmed = trim(source, **TRIM_KWARGS_10HZ)
        expected = _expected_trimmed(data, 1, 2, 76)
        assert bytes(trimmed) == expected

    def test_trim_with_rewindable_audio_reader(self):
        """Rewindable AudioReader should be used directly (streaming)."""
        data = _load_raw(MONO_RAW)
        reader = AudioReader(
            MONO_RAW, sr=SR, sw=SW, ch=1, block_dur=0.1, record=True
        )
        trimmed = trim(reader, **TRIM_KWARGS_10HZ)
        expected = _expected_trimmed(data, 1, 2, 76)
        assert bytes(trimmed) == expected

    def test_trim_with_non_rewindable_audio_reader(self):
        """Non-rewindable AudioReader should be re-wrapped with record=True."""
        data = _load_raw(MONO_RAW)
        reader = AudioReader(
            MONO_RAW, sr=SR, sw=SW, ch=1, block_dur=0.1, record=False
        )
        assert not reader.rewindable
        trimmed = trim(reader, **TRIM_KWARGS_10HZ)
        expected = _expected_trimmed(data, 1, 2, 76)
        assert bytes(trimmed) == expected

    def test_trim_stereo(self):
        data = _load_raw(STEREO_RAW)
        trimmed = trim(STEREO_RAW, sr=SR, sw=SW, ch=2, **TRIM_KWARGS_10HZ)
        expected = _expected_trimmed(data, 2, 2, 76)
        assert bytes(trimmed) == expected

    def test_trim_no_activity_returns_empty_region(self):
        data = _load_raw(MONO_RAW)
        trimmed = trim(
            data,
            sr=SR,
            sw=SW,
            ch=1,
            **{**TRIM_KWARGS_10HZ, "eth": 60},
        )
        assert isinstance(trimmed, AudioRegion)
        assert trimmed.duration == 0.0

    def test_trim_consistent_with_audio_region_trim(self):
        """Module-level trim should give the same result as AudioRegion.trim."""
        data = _load_raw(MONO_RAW)
        region = AudioRegion(data, SR, SW, 1)
        trimmed_method = region.trim(**TRIM_KWARGS_10HZ)
        trimmed_func = trim(data, sr=SR, sw=SW, ch=1, **TRIM_KWARGS_10HZ)
        assert bytes(trimmed_func) == bytes(trimmed_method)

    def test_trim_consistent_across_input_types(self):
        """All input types should produce the same trimmed result."""
        data = _load_raw(STEREO_RAW)
        kw = {**TRIM_KWARGS_10HZ}

        from_file = trim(STEREO_RAW, sr=SR, sw=SW, ch=2, **kw)
        from_path = trim(Path(STEREO_RAW), sr=SR, sw=SW, ch=2, **kw)
        from_bytes = trim(data, sr=SR, sw=SW, ch=2, **kw)
        from_region = trim(AudioRegion(data, SR, SW, 2), **kw)
        from_reader = trim(
            AudioReader(
                STEREO_RAW, sr=SR, sw=SW, ch=2, block_dur=0.1, record=True
            ),
            **kw,
        )
        from_source = trim(
            get_audio_source(STEREO_RAW, sr=SR, sw=SW, ch=2), **kw
        )

        reference = bytes(from_file)
        assert bytes(from_path) == reference
        assert bytes(from_bytes) == reference
        assert bytes(from_region) == reference
        assert bytes(from_reader) == reference
        assert bytes(from_source) == reference


# ── AudioReader wrapping AudioReader ─────────────────────────────────


class TestAudioReaderWrapping:
    """Tests for AudioReader accepting another AudioReader as input."""

    def test_audio_reader_wraps_non_rewindable(self):
        """AudioReader(AudioReader(record=False), record=True) should work."""
        data = _load_raw(MONO_RAW)
        inner = AudioReader(MONO_RAW, sr=SR, sw=SW, ch=1, block_dur=0.1)
        outer = AudioReader(inner, block_dur=0.1, record=True)
        outer.open()
        blocks = []
        while True:
            block = outer.read()
            if block is None:
                break
            blocks.append(block)
        outer.rewind()
        assert outer.data == data

    def test_audio_reader_wraps_rewindable(self):
        """AudioReader(AudioReader(record=True), record=True) should work."""
        data = _load_raw(MONO_RAW)
        inner = AudioReader(
            MONO_RAW, sr=SR, sw=SW, ch=1, block_dur=0.1, record=True
        )
        outer = AudioReader(inner, block_dur=0.1, record=True)
        outer.open()
        blocks = []
        while True:
            block = outer.read()
            if block is None:
                break
            blocks.append(block)
        outer.rewind()
        assert outer.data == data

    def test_wrapped_reader_preserves_audio_params(self):
        """Wrapped reader should expose the same sr, sw, ch."""
        inner = AudioReader(MONO_RAW, sr=SR, sw=SW, ch=1, block_dur=0.1)
        outer = AudioReader(inner, block_dur=0.1, record=True)
        assert outer.sr == SR
        assert outer.sw == SW
        assert outer.ch == 1



import sys
import wave
from functools import partial

import pytest

from auditok import AudioReader, BufferAudioSource, Recorder, WaveAudioSource
from auditok.util import _Limiter, _OverlapAudioReader

AUDIO_FILE_ONE_TO_SIX = "tests/data/1to6arabic_16000_mono_bc_noise.wav"


def _read_all_data(reader):
    blocks = []
    while True:
        data = reader.read()
        if data is None:
            break
        blocks.append(data)
    return b"".join(blocks)


class TestAudioReaderWithFileAudioSource:
    @pytest.fixture(autouse=True)
    def setup_and_teardown(self):
        self.audio_source = WaveAudioSource(filename=AUDIO_FILE_ONE_TO_SIX)
        self.audio_source.open()
        yield
        self.audio_source.close()

    def test_AudioReader_type(self):
        reader = AudioReader(input=self.audio_source)
        err_msg = "wrong object type, expected: 'AudioReader', found: {0}"
        assert isinstance(reader, AudioReader), err_msg.format(type(reader))

    def _test_default_block_size(self):
        reader = AudioReader(input=self.audio_source)
        data = reader.read()
        size = len(data)
        assert (
            size == 160
        ), "Wrong default block_size, expected: 160, found: {0}".format(size)

    @pytest.mark.parametrize(
        "block_dur, expected_nb_samples",
        [
            (None, 160),  # default: 10 ms
            (0.025, 400),  # 25 ms
        ],
        ids=["default", "_25ms"],
    )
    def test_block_duration(self, block_dur, expected_nb_samples):
        """Test the number of samples read for a given block duration."""
        if block_dur is not None:
            reader = AudioReader(input=self.audio_source, block_dur=block_dur)
        else:
            reader = AudioReader(input=self.audio_source)
        data = reader.read()
        nb_samples = len(data) // reader.sample_width
        assert (
            nb_samples == expected_nb_samples
        ), f"Wrong block_size, expected: {expected_nb_samples}, found: {nb_samples}"

    @pytest.mark.parametrize(
        "block_dur, hop_dur, expected_nb_blocks, expected_last_block_nb_samples",
        [
            (None, None, 1879, 126),  # default: 10 ms
            (0.01, None, 1879, 126),  # block_dur_10ms_hop_dur_None
            (0.01, 0.01, 1879, 126),  # block_dur_10ms_hop_dur_10ms
            (0.02, None, 940, 126),  # block_dur_20ms_hop_dur_None
            (0.025, None, 752, 206),  # block_dur_25ms_hop_dur_None
            (0.02, 0.01, 1878, 286),  # block_dur_20ms_hop_dur_10ms
            (0.025, 0.005, 3754, 366),  # block_dur_25ms_hop_dur_5ms
        ],
        ids=[
            "default",
            "block_dur_10ms_hop_dur_None",
            "block_dur_10ms_hop_dur_10ms",
            "block_dur_20ms_hop_dur_None",
            "block_dur_25ms_hop_dur_None",
            "block_dur_20ms_hop_dur_10ms",
            "block_dur_25ms_hop_dur_5ms",
        ],
    )
    def test_hop_duration(
        self,
        block_dur,
        hop_dur,
        expected_nb_blocks,
        expected_last_block_nb_samples,
    ):
        """Test the number of read blocks and the duration of last block for
        different 'block_dur' and 'hop_dur' values.

        Args:
            block_dur (float or None): block duration in seconds.
            hop_dur (float or None): hop duration in seconds.
            expected_nb_blocks (int): expected number of read block.
            expected_last_block_nb_samples (int): expected number of sample
                in the last block.
        """
        if block_dur is not None:
            reader = AudioReader(
                input=self.audio_source, block_dur=block_dur, hop_dur=hop_dur
            )
        else:
            reader = AudioReader(input=self.audio_source, hop_dur=hop_dur)

        nb_blocks = 0
        last_block_nb_samples = None
        while True:
            data = reader.read()
            if data is not None:
                nb_blocks += 1
                last_block_nb_samples = len(data) // reader.sample_width
            else:
                break
        err_msg = "Wrong number of blocks read from source, expected: "
        err_msg += f"{expected_nb_blocks}, found: {nb_blocks}"
        assert nb_blocks == expected_nb_blocks, err_msg

        err_msg = (
            "Wrong number of samples in last block read from source, expected: "
        )
        err_msg += (
            f"{expected_last_block_nb_samples}, found: {last_block_nb_samples}"
        )

        assert last_block_nb_samples == expected_last_block_nb_samples, err_msg

    def test_hop_duration_exception(self):
        """Test passing hop_dur > block_dur raises ValueError"""
        with pytest.raises(ValueError):
            AudioReader(self.audio_source, block_dur=0.01, hop_dur=0.015)

    def test_hop_dur_same_sample_count_as_block_dur(self):
        """Test that hop_dur and block_dur yielding the same sample count
        after int truncation raises ValueError.

        With sr=16000: int(0.00012 * 16000) = int(1.92) = 1
                       int(0.00007 * 16000) = int(1.12) = 1
        """
        with pytest.raises(ValueError, match="hop_size.*>=.*block_size"):
            AudioReader(self.audio_source, block_dur=0.00012, hop_dur=0.00007)

    def test_hop_dur_zero_samples(self):
        """Test that hop_dur too small to cover one sample raises ValueError.

        With sr=16000: int(0.00005 * 16000) = int(0.8) = 0
        """
        with pytest.raises(ValueError, match="too small"):
            AudioReader(self.audio_source, block_dur=0.001, hop_dur=0.00005)

    def test_block_dur_zero_samples(self):
        """Test that block_dur too small to cover one sample raises
        TooSmallBlockDuration (a ValueError subclass).

        With sr=16000: int(0.00005 * 16000) = int(0.8) = 0
        """
        with pytest.raises(ValueError, match="Too small block_dur"):
            AudioReader(self.audio_source, block_dur=0.00005)

    @pytest.mark.parametrize(
        "block_dur, hop_dur",
        [
            (None, None),  # default
            (0.01, None),  # block_dur_10ms_hop_dur_None
            (None, 0.01),  # block_dur_None__hop_dur_10ms
            (0.05, 0.05),  # block_dur_50ms_hop_dur_50ms
        ],
        ids=[
            "default",
            "block_dur_10ms_hop_dur_None",
            "block_dur_None__hop_dur_10ms",
            "block_dur_50ms_hop_dur_50ms",
        ],
    )
    def test_reader_class_block_dur_equals_hop_dur(self, block_dur, hop_dur):
        """Test passing hop_dur == block_dur does not create an instance of
        '_OverlapAudioReader'.
        """
        if block_dur is not None:
            reader = AudioReader(
                input=self.audio_source, block_dur=block_dur, hop_dur=hop_dur
            )
        else:
            reader = AudioReader(input=self.audio_source, hop_dur=hop_dur)
        assert not isinstance(reader, _OverlapAudioReader)

    def test_sampling_rate(self):
        reader = AudioReader(input=self.audio_source)
        sampling_rate = reader.sampling_rate
        assert (
            sampling_rate == 16000
        ), f"Wrong sampling rate, expected: 16000, found: {sampling_rate}"

    def test_sample_width(self):
        reader = AudioReader(input=self.audio_source)
        sample_width = reader.sample_width
        assert (
            sample_width == 2
        ), f"Wrong sample width, expected: 2, found: {sample_width}"

    def test_channels(self):
        reader = AudioReader(input=self.audio_source)
        channels = reader.channels
        assert (
            channels == 1
        ), f"Wrong number of channels, expected: 1, found: {channels}"

    def test_read(self):
        reader = AudioReader(input=self.audio_source, block_dur=0.02)
        reader_data = reader.read()
        audio_source = WaveAudioSource(filename=AUDIO_FILE_ONE_TO_SIX)
        audio_source.open()
        audio_source_data = audio_source.read(320)
        audio_source.close()
        assert (
            reader_data == audio_source_data
        ), "Unexpected data read from AudioReader"

    def test_read_with_overlap(self):
        reader = AudioReader(
            input=self.audio_source, block_dur=0.02, hop_dur=0.01
        )
        _ = reader.read()  # first block
        reader_data = reader.read()  # second block with 0.01 S overlap
        audio_source = WaveAudioSource(filename=AUDIO_FILE_ONE_TO_SIX)
        audio_source.open()
        _ = audio_source.read(160)
        audio_source_data = audio_source.read(320)
        audio_source.close()
        assert (
            reader_data == audio_source_data
        ), "Unexpected data read from AudioReader"

    def test_read_from_AudioReader_with_max_read(self):
        # read a maximum of 0.75 seconds from audio source
        reader = AudioReader(input=self.audio_source, max_read=0.75)
        assert isinstance(reader._audio_source._audio_source, _Limiter)
        reader_data = _read_all_data(reader)

        audio_source = WaveAudioSource(filename=AUDIO_FILE_ONE_TO_SIX)
        audio_source.open()
        audio_source_data = audio_source.read(int(16000 * 0.75))
        audio_source.close()

        assert (
            reader_data == audio_source_data
        ), f"Unexpected data read from AudioReader with 'max_read = {0.75}'"

    def test_read_data_size_from_AudioReader_with_max_read(self):
        # read a maximum of 1.191 seconds from audio source
        reader = AudioReader(input=self.audio_source, max_read=1.191)
        assert isinstance(reader._audio_source._audio_source, _Limiter)
        total_samples = round(reader.sampling_rate * 1.191)
        block_size = int(reader.block_dur * reader.sampling_rate)
        nb_full_blocks, last_block_size = divmod(total_samples, block_size)
        total_samples_with_overlap = (
            nb_full_blocks * block_size + last_block_size
        )
        expected_read_bytes = (
            total_samples_with_overlap * reader.sample_width * reader.channels
        )

        reader_data = _read_all_data(reader)
        total_read = len(reader_data)
        err_msg = (
            "Wrong data length read from LimiterADS, "
            f"expected: {expected_read_bytes}, found: {total_read}"
        )
        assert total_read == expected_read_bytes, err_msg

    def test_read_from_Recorder(self):
        reader = Recorder(input=self.audio_source, block_dur=0.025)
        reader_data = []
        for _ in range(10):
            block = reader.read()
            if block is None:
                break
            reader_data.append(block)
        reader_data = b"".join(reader_data)

        audio_source = WaveAudioSource(filename=AUDIO_FILE_ONE_TO_SIX)
        audio_source.open()
        audio_source_data = audio_source.read(400 * 10)
        audio_source.close()

        assert (
            reader_data == audio_source_data
        ), "Unexpected data read from Recorder"

    def test_AudioReader_rewindable(self):
        reader = AudioReader(input=self.audio_source, record=True)
        assert (
            reader.rewindable
        ), "AudioReader with record=True should be rewindable"

    def test_AudioReader_record_and_rewind(self):
        reader = AudioReader(
            input=self.audio_source, record=True, block_dur=0.02
        )
        # read 0.02 * 10 = 0.2 sec. of data
        for _ in range(10):
            reader.read()
        reader.rewind()

        # read all available data after rewind
        reader_data = _read_all_data(reader)

        audio_source = WaveAudioSource(filename=AUDIO_FILE_ONE_TO_SIX)
        audio_source.open()
        audio_source_data = audio_source.read(320 * 10)  # read 0.2 sec. of data
        audio_source.close()

        assert (
            reader_data == audio_source_data
        ), "Unexpected data read from AudioReader with record = True"

    def test_Recorder_record_and_rewind(self):
        recorder = Recorder(input=self.audio_source, block_dur=0.02)
        # read 0.02 * 10 = 0.2 sec. of data
        for _ in range(10):
            recorder.read()

        recorder.rewind()

        # read all available data after rewind
        recorder_data = []
        recorder_data = _read_all_data(recorder)

        audio_source = WaveAudioSource(filename=AUDIO_FILE_ONE_TO_SIX)
        audio_source.open()
        audio_source_data = audio_source.read(320 * 10)  # read 0.2 sec. of data
        audio_source.close()

        assert (
            recorder_data == audio_source_data
        ), "Unexpected data read from Recorder"

    def test_read_overlapping_blocks(self):
        # Use arbitrary valid block_size and hop_size
        block_size = 1714
        hop_size = 313
        block_dur = block_size / self.audio_source.sampling_rate
        hop_dur = hop_size / self.audio_source.sampling_rate

        reader = AudioReader(
            input=self.audio_source,
            block_dur=block_dur,
            hop_dur=hop_dur,
        )

        # Read all available overlapping blocks of data
        reader_data = []
        while True:
            block = reader.read()
            if block is None:
                break
            reader_data.append(block)

        # Read all data from file and build a BufferAudioSource
        fp = wave.open(AUDIO_FILE_ONE_TO_SIX, "r")
        wave_data = fp.readframes(fp.getnframes())
        fp.close()
        audio_source = BufferAudioSource(
            wave_data,
            reader.sampling_rate,
            reader.sample_width,
            reader.channels,
        )
        audio_source.open()

        # Compare all blocks read from OverlapADS to those read from an
        # audio source with a manual position setting
        for i, block in enumerate(reader_data):
            tmp = audio_source.read(block_size)
            assert (
                block == tmp
            ), f"Unexpected data (block {i}) from reader with overlapping blocks"
            audio_source.position = (i + 1) * hop_size

        audio_source.close()

    def test_read_overlapping_blocks_with_max_read(self):
        block_size = 256
        hop_size = 200
        block_dur = block_size / self.audio_source.sampling_rate
        hop_dur = hop_size / self.audio_source.sampling_rate

        reader = AudioReader(
            input=self.audio_source,
            block_dur=block_dur,
            hop_dur=hop_dur,
            max_read=0.5,
        )

        # Read all available overlapping blocks of data
        reader_data = []
        while True:
            block = reader.read()
            if block is None:
                break
            reader_data.append(block)

        # Read all data from file and build a BufferAudioSource
        fp = wave.open(AUDIO_FILE_ONE_TO_SIX, "r")
        wave_data = fp.readframes(fp.getnframes())
        fp.close()
        audio_source = BufferAudioSource(
            wave_data,
            reader.sampling_rate,
            reader.sample_width,
            reader.channels,
        )
        audio_source.open()

        # Compare all blocks read from OverlapADS to those read from an
        # audio source with a manual position setting
        for i, block in enumerate(reader_data):
            tmp = audio_source.read(len(block) // (reader.sw * reader.ch))
            assert block == tmp, (
                f"Unexpected data (block {i}) from reader"
                " with overlapping blocks and max_read"
            )
            audio_source.position = (i + 1) * hop_size

        audio_source.close()

    def test_length_read_overlapping_blocks_with_max_read(self):
        block_size = 313
        hop_size = 207
        block_dur = block_size / self.audio_source.sampling_rate
        hop_dur = hop_size / self.audio_source.sampling_rate

        reader = AudioReader(
            input=self.audio_source,
            max_read=1.932,
            block_dur=block_dur,
            hop_dur=hop_dur,
        )

        total_samples = round(reader.sampling_rate * 1.932)
        first_read_size = block_size
        next_read_size = block_size - hop_size
        nb_next_blocks, last_block_size = divmod(
            (total_samples - first_read_size), next_read_size
        )
        total_samples_with_overlap = (
            first_read_size + next_read_size * nb_next_blocks + last_block_size
        )
        expected_read_bytes = (
            total_samples_with_overlap * reader.sw * reader.channels
        )

        cache_size = (
            (block_size - hop_size) * reader.sample_width * reader.channels
        )
        total_read = cache_size

        i = 0
        while True:
            block = reader.read()
            if block is None:
                break
            i += 1
            total_read += len(block) - cache_size

        err_msg = (
            "Wrong data length read from LimiterADS, expected: {0}, found: {1}"
        )
        assert total_read == expected_read_bytes, err_msg.format(
            expected_read_bytes, total_read
        )

    def test_reader_with_overlapping_blocks__rewindable(self):
        reader = AudioReader(
            input=self.audio_source,
            block_dur=320,
            hop_dur=160,
            record=True,
        )
        assert (
            reader.rewindable
        ), "AudioReader with record=True should be rewindable"

    def test_overlapping_blocks_with_max_read_rewind_and_read(self):
        # Use arbitrary valid block_size and hop_size
        block_size = 1600
        hop_size = 400
        block_dur = block_size / self.audio_source.sampling_rate
        hop_dur = hop_size / self.audio_source.sampling_rate

        reader = AudioReader(
            input=self.audio_source,
            block_dur=block_dur,
            hop_dur=hop_dur,
            record=True,
        )

        # Read all available data overlapping blocks
        i = 0
        while True:
            block = reader.read()
            if block is None:
                break
            i += 1

        reader.rewind()

        # Read all data from file and build a BufferAudioSource
        fp = wave.open(AUDIO_FILE_ONE_TO_SIX, "r")
        wave_data = fp.readframes(fp.getnframes())
        fp.close()
        audio_source = BufferAudioSource(
            wave_data,
            reader.sampling_rate,
            reader.sample_width,
            reader.channels,
        )
        audio_source.open()

        # Compare blocks read from AudioReader to those read from
        # an BufferAudioSource with manual position setting
        for j in range(i):
            tmp = audio_source.read(block_size)
            assert reader.read() == tmp, (
                f"Unexpected data (block {i}) from reader"
                " with overlapping blocks and record = True"
            )
            audio_source.position = (j + 1) * hop_size

        audio_source.close()

    def test_overlapping_blocks_with_record_and_max_read_rewind_and_read(self):
        # Use arbitrary valid block_size and hop_size
        block_size = 1600
        hop_size = 400
        block_dur = block_size / self.audio_source.sampling_rate
        hop_dur = hop_size / self.audio_source.sampling_rate

        reader = AudioReader(
            input=self.audio_source,
            max_time=1.50,
            block_dur=block_dur,
            hop_dur=hop_dur,
            record=True,
        )

        # Read all available data overlapping blocks
        i = 0
        while True:
            block = reader.read()
            if block is None:
                break
            i += 1

        reader.rewind()

        # Read all data from file and build a BufferAudioSource
        fp = wave.open(AUDIO_FILE_ONE_TO_SIX, "r")
        wave_data = fp.readframes(fp.getnframes())
        fp.close()
        audio_source = BufferAudioSource(
            wave_data,
            reader.sampling_rate,
            reader.sample_width,
            reader.channels,
        )
        audio_source.open()

        # Compare all blocks read from AudioReader to those read
        # from BufferAudioSource with a manual position setting
        for j in range(i):
            tmp = audio_source.read(block_size)
            assert (
                reader.read() == tmp
            ), "Unexpected block (N={0}) read from OverlapADS".format(i)
            audio_source.position = (j + 1) * hop_size

        audio_source.close()

    def test_length_read_overlapping_blocks_with_record_and_max_read(self):
        # Use arbitrary valid block_size and hop_size
        block_size = 1000
        hop_size = 200
        block_dur = block_size / self.audio_source.sampling_rate
        hop_dur = hop_size / self.audio_source.sampling_rate

        reader = AudioReader(
            input=self.audio_source,
            block_dur=block_dur,
            hop_dur=hop_dur,
            record=True,
            max_read=1.317,
        )
        total_samples = round(reader.sampling_rate * 1.317)
        first_read_size = block_size
        next_read_size = block_size - hop_size
        nb_next_blocks, last_block_size = divmod(
            (total_samples - first_read_size), next_read_size
        )
        total_samples_with_overlap = (
            first_read_size + next_read_size * nb_next_blocks + last_block_size
        )
        expected_read_bytes = (
            total_samples_with_overlap * reader.sample_width * reader.channels
        )

        cache_size = (
            (block_size - hop_size) * reader.sample_width * reader.channels
        )
        total_read = cache_size

        i = 0
        while True:
            block = reader.read()
            if block is None:
                break
            i += 1
            total_read += len(block) - cache_size

        err_msg = (
            "Wrong data length read from AudioReader, "
            f"expected: {expected_read_bytes}, found: {total_read}"
        )
        assert total_read == expected_read_bytes, err_msg


def test_AudioReader_raw_data():

    data = b"ABCDEFGHIJKLMNOPQRSTUVWXYZ012345"
    block_size = 5
    hop_size = 4
    reader = AudioReader(
        input=data,
        sampling_rate=16,
        sample_width=2,
        channels=1,
        block_dur=block_size / 16,
        hop_dur=hop_size / 16,
        max_read=0.80,
        record=True,
    )
    reader.open()

    assert (
        reader.sampling_rate == 16
    ), f"Wrong sampling rate, expected: 16, found: {reader.sampling_rate}"

    assert (
        reader.sample_width == 2
    ), f"Wrong sample width, expected: 2, found: {reader.sample_width}"

    # Read all available data overlapping blocks
    i = 0
    while True:
        block = reader.read()
        if block is None:
            break
        i += 1

    reader.rewind()

    # Build a BufferAudioSource
    audio_source = BufferAudioSource(
        data, reader.sampling_rate, reader.sample_width, reader.channels
    )
    audio_source.open()

    # Compare all blocks read from AudioReader to those read from an audio
    # source with a manual position setting
    for j in range(i):
        tmp = audio_source.read(block_size)
        block = reader.read()
        assert (
            block == tmp
        ), f"Unexpected block {block!r} (N={i}) read from OverlapADS"
        audio_source.position = (j + 1) * hop_size
    audio_source.close()
    reader.close()


def test_hop_dur_same_sample_count_as_block_dur_low_sr():
    """Test with sr=8000 where realistic durations collide.

    int(0.0011 * 8000) = int(8.8) = 8
    int(0.0010 * 8000) = int(8.0) = 8
    """
    data = b"\x00" * 160  # 10ms at sr=8000, sw=1, ch=1
    source = BufferAudioSource(
        data, sampling_rate=8000, sample_width=1, channels=1
    )
    source.open()
    with pytest.raises(ValueError, match="hop_size.*>=.*block_size"):
        AudioReader(source, block_dur=0.0011, hop_dur=0.0010)
    source.close()


def test_hop_dur_close_but_different_sample_count():
    """Verify no error when hop_dur and block_dur are close but yield
    different sample counts.

    With sr=16000: int(0.00019 * 16000) = int(3.04) = 3
                   int(0.00013 * 16000) = int(2.08) = 2
    """
    data = b"\x00" * 320  # some raw audio
    source = BufferAudioSource(
        data, sampling_rate=16000, sample_width=1, channels=1
    )
    source.open()
    reader = AudioReader(source, block_dur=0.00019, hop_dur=0.00013)
    assert isinstance(reader._audio_source, _OverlapAudioReader)
    source.close()


def test_AudioReader_alias_params():
    reader = AudioReader(
        input=b"0" * 1600,
        sr=16000,
        sw=2,
        channels=1,
    )
    assert reader.sampling_rate == 16000, (
        "Unexpected sampling rate: reader.sampling_rate = "
        + f"{reader.sampling_rate} instead of 16000"
    )
    assert reader.sr == 16000, (
        "Unexpected sampling rate: reader.sr = "
        + f"{reader.sr} instead of 16000"
    )
    assert reader.sample_width == 2, (
        "Unexpected sample width: reader.sample_width = "
        + f"{reader.sample_width} instead of 2"
    )
    assert reader.sw == 2, (
        "Unexpected sample width: reader.sw = " + f"{reader.sw} instead of 2"
    )
    assert reader.channels == 1, (
        "Unexpected number of channels: reader.channels = "
        + f"{reader.channels} instead of 1"
    )
    assert reader.ch == 1, (
        "Unexpected number of channels: reader.ch = "
        + f"{reader.ch} instead of 1"
    )


@pytest.mark.parametrize(
    "file_id, max_read, size",
    [
        ("mono_400", 0.5, 16000),  # mono
        ("3channel_400-800-1600", 0.5, 16000 * 3),  # multichannel
    ],
    ids=["mono", "multichannel"],
)
def test_Limiter(file_id, max_read, size):
    input_wav = "tests/data/test_16KHZ_{}Hz.wav".format(file_id)
    input_raw = "tests/data/test_16KHZ_{}Hz.raw".format(file_id)
    with open(input_raw, "rb") as fp:
        expected = fp.read(size)

    reader = AudioReader(input_wav, block_dur=0.1, max_read=max_read)
    reader.open()
    data = _read_all_data(reader)
    reader.close()
    assert data == expected


@pytest.mark.parametrize(
    "file_id",
    [
        "mono_400",  # mono
        "3channel_400-800-1600",  # multichannel
    ],
    ids=["mono", "multichannel"],
)
def test_Recorder(file_id):
    input_wav = "tests/data/test_16KHZ_{}Hz.wav".format(file_id)
    input_raw = "tests/data/test_16KHZ_{}Hz.raw".format(file_id)
    with open(input_raw, "rb") as fp:
        expected = fp.read()

    reader = AudioReader(input_wav, block_dur=0.1, record=True)
    reader.open()
    data = _read_all_data(reader)
    assert data == expected

    # rewind many times
    for _ in range(3):
        reader.rewind()
        data = _read_all_data(reader)
        assert data == expected
        assert data == reader.data
    reader.close()


@pytest.mark.parametrize(
    "file_id",
    [
        "mono_400",  # mono
        "3channel_400-800-1600",  # multichannel
    ],
    ids=["mono", "multichannel"],
)
def test_Recorder_alias(file_id):
    input_wav = "tests/data/test_16KHZ_{}Hz.wav".format(file_id)
    input_raw = "tests/data/test_16KHZ_{}Hz.raw".format(file_id)
    with open(input_raw, "rb") as fp:
        expected = fp.read()

    reader = Recorder(input_wav, block_dur=0.1)
    reader.open()
    data = _read_all_data(reader)
    assert data == expected

    # rewind many times
    for _ in range(3):
        reader.rewind()
        data = _read_all_data(reader)
        assert data == expected
        assert data == reader.data
    reader.close()



"""Interactive audio visualization widget for Jupyter notebooks.

Renders an HTML5 Canvas waveform with clickable detection regions and
Web Audio API playback.  The widget is entirely self-contained (no
external JS libraries) and is injected into the notebook output cell
via ``IPython.display.display()``.
"""

import base64
import io
import json
import uuid
import wave

import numpy as np

_RULER_H = 25

_WIDGET_HTML = """\
<style>
  #{wid} .atk-btn {{
    border:none; border-radius:4px; padding:4px 12px;
    cursor:pointer; font-weight:bold; font-size:12px;
    display:inline-flex; align-items:center; gap:5px;
    transition: filter 0.2s ease;
  }}
  #{wid} .atk-btn:hover {{ filter:brightness(1.15); }}
  #{wid} .atk-btn:active {{ filter:brightness(0.85); }}
  #{wid} .atk-btn svg {{ vertical-align:middle; }}
</style>
<div id="{wid}" style="
    background:{bg}; border-radius:6px; padding:10px; margin:6px 0;
    font-family:monospace; font-size:12px; color:#ccc;
    box-sizing:border-box; width:100%;
">
  <div style="margin-bottom:6px">
    <b style="color:#40d970">auditok</b>
    <span style="margin-left:8px">{title}</span>
  </div>
  <canvas id="{wid}_cv"
    style="cursor:pointer; display:block; border-radius:4px; width:100%; height:{height}px">
  </canvas>
  <div style="margin-top:6px; display:flex; align-items:center; gap:10px">
    <button id="{wid}_play" class="atk-btn" style="
        background:#40d970; color:#282a36; min-width:90px; justify-content:center;
    "><svg width="12" height="12" viewBox="0 0 12 12"><polygon points="2,0 2,12 11,6" fill="currentColor"/></svg> Play all</button>
    <button id="{wid}_stop" class="atk-btn" style="
        background:#e31f8f; color:#fff;
    "><svg width="12" height="12" viewBox="0 0 12 12"><rect x="1" y="1" width="10" height="10" rx="1" fill="currentColor"/></svg> Stop</button>
    <span id="{wid}_time" style="
        color:#40d970; font-size:12px; font-weight:bold;
        user-select:all; -webkit-user-select:all; cursor:text;
    "></span>
    <span id="{wid}_info" style="color:#999; font-size:11px"></span>
  </div>
</div>
<script>
(function() {{
  var W = document.getElementById("{wid}");
  var cv = document.getElementById("{wid}_cv");
  var ctx = cv.getContext("2d");
  var info = document.getElementById("{wid}_info");
  var btnPlay = document.getElementById("{wid}_play");
  var btnStop = document.getElementById("{wid}_stop");
  var timeSpan = document.getElementById("{wid}_time");

  var sr = {sr};
  var nch = {nch};
  var duration = {duration};
  var peaks = {peaks_json};
  var detections = {detections_json};
  var regionB64 = {regions_b64_json};
  var ethNorm = {eth_json};  // normalised energy threshold per channel (or null)

  var RULER_H = {ruler_h};
  var audioCtx = null;
  var currentSource = null;
  var fullAudioB64 = "{full_audio_b64}";
  var fullAudioBuf = null;  // cached decoded AudioBuffer for seeking
  var hoveredRegion = -1;
  var selectedRegion = -1;
  var rulerHoverX = -1;     // x pixel when hovering ruler, -1 if not
  var animId = null;
  var playStartTime = 0;
  var playOffset = 0;
  var playDuration = 0;
  var playGen = 0;          // generation counter to cancel stale async decodes
  var playing = false;      // true while audio is actively playing
  var pausedAt = -1;        // signal time where playback was paused (-1 = not paused)
  var pausedLabel = "";     // info label to restore on resume
  var pxPerSec = 1;

  var dpr = window.devicePixelRatio || 1;
  var cssW = 0;   // logical width in CSS pixels (all drawing uses this)
  var cssH = {height};

  function sizeCanvas() {{
    var w = cv.clientWidth || cv.parentElement.clientWidth;
    if (w <= 0) return;
    cssW = w;
    // Buffer at device resolution, CSS pins the display size
    cv.width = cssW * dpr;
    cv.height = cssH * dpr;
    cv.style.height = cssH + "px";
    ctx.setTransform(dpr, 0, 0, dpr, 0, 0);  // scale once, draw in CSS px
    pxPerSec = cssW / duration;
  }}
  sizeCanvas();

  function b64ToArrayBuffer(b64) {{
    var bin = atob(b64);
    var buf = new Uint8Array(bin.length);
    for (var i = 0; i < bin.length; i++) buf[i] = bin.charCodeAt(i);
    return buf.buffer;
  }}

  // Choose a "nice" tick interval so ticks are readable but not crowded.
  function tickInterval() {{
    var nice = [0.01, 0.02, 0.05, 0.1, 0.2, 0.5,
                1, 2, 5, 10, 15, 30, 60, 120, 300, 600];
    var target = duration / (cssW / 80);  // ~80px between ticks
    for (var i = 0; i < nice.length; i++) {{
      if (nice[i] >= target) return nice[i];
    }}
    return nice[nice.length - 1];
  }}

  function formatTime(t) {{
    if (duration < 60) return t.toFixed(2) + "s";
    if (duration < 3600) {{
      var m = Math.floor(t / 60);
      var s = (t % 60).toFixed(1);
      return m + ":" + (s < 10 ? "0" : "") + s;
    }}
    var h = Math.floor(t / 3600);
    var m = Math.floor((t % 3600) / 60);
    var s = Math.floor(t % 60);
    return h + ":" + (m < 10 ? "0" : "") + m + ":" + (s < 10 ? "0" : "") + s;
  }}

  function drawRuler() {{
    var w = cssW;
    // Ruler background
    ctx.fillStyle = "#1e1a2e";
    ctx.fillRect(0, 0, w, RULER_H);
    // Bottom border
    ctx.strokeStyle = "#555";
    ctx.lineWidth = 1;
    ctx.beginPath();
    ctx.moveTo(0, snap(RULER_H - 1));
    ctx.lineTo(w, snap(RULER_H - 1));
    ctx.stroke();

    var interval = tickInterval();
    ctx.fillStyle = "#999";
    ctx.strokeStyle = "#666";
    ctx.font = "10px monospace";
    ctx.textAlign = "center";
    ctx.lineWidth = 1;

    var t = 0;
    while (t <= duration) {{
      var x = snap(t * pxPerSec);
      // Tick mark
      ctx.beginPath();
      ctx.moveTo(x, RULER_H - 1);
      ctx.lineTo(x, RULER_H - 7);
      ctx.stroke();
      // Label (skip if too close to the edge)
      if (x > 20 && x < w - 20) {{
        ctx.fillText(formatTime(t), x, RULER_H - 9);
      }}
      t += interval;
    }}
  }}

  function drawNeedle(x) {{
    // Downward triangle at the ruler bottom
    var sz = 5;
    ctx.fillStyle = "#e31f8f";
    ctx.beginPath();
    ctx.moveTo(x - sz, RULER_H - 1);
    ctx.lineTo(x + sz, RULER_H - 1);
    ctx.lineTo(x, RULER_H + sz);
    ctx.closePath();
    ctx.fill();
    // Vertical dashed line through the waveform area
    ctx.strokeStyle = "rgba(227,31,143,0.4)";
    ctx.lineWidth = 1;
    ctx.setLineDash([4, 4]);
    ctx.beginPath();
    ctx.moveTo(x, RULER_H);
    ctx.lineTo(x, cssH);
    ctx.stroke();
    ctx.setLineDash([]);
  }}

  // Snap to nearest half-pixel for crisp 1px lines
  function snap(v) {{ return Math.round(v) + 0.5; }}

  function drawWaveform() {{
    var w = cssW, h = cssH;
    var waveH = h - RULER_H;
    var chH = waveH / nch;

    ctx.fillStyle = "{plot_bg}";
    ctx.fillRect(0, 0, w, h);

    // Draw detections background (waveform area only)
    for (var d = 0; d < detections.length; d++) {{
      var det = detections[d];
      var x0 = det[0] * pxPerSec;
      var x1 = det[1] * pxPerSec;
      var isHovered = (d === hoveredRegion);
      var isSelected = (d === selectedRegion);
      ctx.fillStyle = isSelected ? "rgba(255,140,26,0.45)"
                     : isHovered ? "rgba(255,140,26,0.30)"
                     : "rgba(119,119,119,0.5)";
      ctx.fillRect(x0, RULER_H, x1 - x0, waveH);

      if (isSelected || isHovered) {{
        ctx.strokeStyle = "#ff8c1a";
        ctx.lineWidth = isSelected ? 2 : 1;
        ctx.strokeRect(Math.round(x0), RULER_H, Math.round(x1 - x0), waveH);
      }}
    }}

    // Draw waveform per channel
    for (var ch = 0; ch < nch; ch++) {{
      var chPeaks = peaks[ch];
      var mid = RULER_H + chH * ch + chH / 2;
      ctx.strokeStyle = "{signal_color}";
      ctx.lineWidth = 1;
      ctx.beginPath();
      for (var i = 0; i < chPeaks.length; i++) {{
        var x = snap((i / chPeaks.length) * w);
        var minV = chPeaks[i][0];
        var maxV = chPeaks[i][1];
        ctx.moveTo(x, snap(mid - maxV * chH * 0.45));
        ctx.lineTo(x, snap(mid - minV * chH * 0.45));
      }}
      ctx.stroke();

      // Energy threshold line
      if (ethNorm[ch] != null) {{
        var thY = snap(mid - ethNorm[ch] * chH * 0.45);
        ctx.strokeStyle = "#e31f8f";
        ctx.lineWidth = 1;
        ctx.setLineDash([6, 4]);
        ctx.beginPath();
        ctx.moveTo(0, thY);
        ctx.lineTo(w, thY);
        ctx.stroke();
        ctx.setLineDash([]);
      }}

      // Channel separator
      if (ch > 0) {{
        ctx.strokeStyle = "#555";
        ctx.lineWidth = 0.5;
        ctx.beginPath();
        ctx.moveTo(0, snap(RULER_H + chH * ch));
        ctx.lineTo(w, snap(RULER_H + chH * ch));
        ctx.stroke();
      }}
    }}

    // Detection labels
    ctx.font = "10px monospace";
    ctx.textAlign = "center";
    for (var d = 0; d < detections.length; d++) {{
      var det = detections[d];
      var x0 = det[0] * pxPerSec;
      var x1 = det[1] * pxPerSec;
      if (x1 - x0 > 18) {{
        ctx.fillStyle = (d === hoveredRegion || d === selectedRegion)
            ? "#ff8c1a" : "#ccc";
        ctx.fillText(d + 1, (x0 + x1) / 2, RULER_H + 12);
      }}
    }}

    // Ruler on top (drawn last so it's above detection highlights)
    drawRuler();

    // Needle cursor when hovering ruler
    if (rulerHoverX >= 0) drawNeedle(rulerHoverX);
  }}

  function drawPlayhead(t) {{
    var x = snap(t * pxPerSec);
    ctx.strokeStyle = "#e31f8f";
    ctx.lineWidth = 1;
    ctx.beginPath();
    ctx.moveTo(x, RULER_H);
    ctx.lineTo(x, cssH);
    ctx.stroke();
    // Small triangle at ruler bottom
    var sz = 4;
    ctx.fillStyle = "#e31f8f";
    ctx.beginPath();
    ctx.moveTo(x - sz, RULER_H);
    ctx.lineTo(x + sz, RULER_H);
    ctx.lineTo(x, RULER_H + sz);
    ctx.closePath();
    ctx.fill();
  }}

  function formatTimePrecise(t) {{
    var m = Math.floor(t / 60);
    var s = t % 60;
    return (m > 0 ? m + ":" : "") + (m > 0 && s < 10 ? "0" : "") + s.toFixed(3) + "s";
  }}

  var ICO_PLAY = '<svg width="12" height="12" viewBox="0 0 12 12"><polygon points="2,0 2,12 11,6" fill="currentColor"/></svg>';
  var ICO_PAUSE = '<svg width="12" height="12" viewBox="0 0 12 12"><rect x="1" y="0" width="3.5" height="12" rx="0.5" fill="currentColor"/><rect x="7.5" y="0" width="3.5" height="12" rx="0.5" fill="currentColor"/></svg>';

  function setPlaying(on) {{
    playing = on;
    btnPlay.innerHTML = on ? ICO_PAUSE + " Pause" : ICO_PLAY + " Play all";
  }}

  function updateTimestamp(t) {{
    timeSpan.textContent = formatTimePrecise(t);
  }}

  function animatePlayhead() {{
    if (!audioCtx || !currentSource) return;
    var elapsed = audioCtx.currentTime - playStartTime;
    if (elapsed > playDuration) {{
      stopPlayback();
      return;
    }}
    var curTime = playOffset + elapsed;
    drawWaveform();
    drawPlayhead(curTime);
    updateTimestamp(curTime);
    animId = requestAnimationFrame(animatePlayhead);
  }}

  function getAudioCtx() {{
    if (!audioCtx) audioCtx = new (window.AudioContext || window.webkitAudioContext)();
    return audioCtx;
  }}

  function stopPlayback() {{
    playGen++;  // invalidate any pending async decode callbacks
    if (currentSource) {{
      currentSource.onended = null;  // prevent stale callback from clobbering next source
      try {{ currentSource.stop(); }} catch(e) {{}}
      currentSource = null;
    }}
    if (animId) {{
      cancelAnimationFrame(animId);
      animId = null;
    }}
    setPlaying(false);
    pausedAt = -1;
    pausedLabel = "";
    info.textContent = "";
    timeSpan.textContent = "";
    drawWaveform();
  }}

  function pausePlayback() {{
    if (!playing || !currentSource) return;
    var elapsed = audioCtx.currentTime - playStartTime;
    pausedAt = playOffset + elapsed;
    pausedLabel = info.textContent;
    playGen++;
    currentSource.onended = null;
    try {{ currentSource.stop(); }} catch(e) {{}}
    currentSource = null;
    if (animId) {{
      cancelAnimationFrame(animId);
      animId = null;
    }}
    setPlaying(false);
    // Keep timestamp and playhead visible at paused position
    drawWaveform();
    drawPlayhead(pausedAt);
    updateTimestamp(pausedAt);
  }}

  function destroy() {{
    stopPlayback();
    if (audioCtx) {{
      try {{ audioCtx.close(); }} catch(e) {{}}
      audioCtx = null;
    }}
    fullAudioBuf = null;
  }}

  // Play a region from its base64 WAV (used for detection click-to-play)
  function playB64(b64, offset, dur, label) {{
    stopPlayback();
    var gen = playGen;
    var actx = getAudioCtx();
    var buf = b64ToArrayBuffer(b64);
    actx.decodeAudioData(buf, function(audioBuf) {{
      if (gen !== playGen) return;  // superseded by another play/stop
      _startSource(audioBuf, 0, audioBuf.duration, offset, label);
    }});
  }}

  // Play the full audio from a given time offset (used for ruler seek)
  function playFullFrom(seekTime) {{
    stopPlayback();
    var gen = playGen;
    var actx = getAudioCtx();

    function _play(audioBuf) {{
      if (gen !== playGen) return;  // superseded by another play/stop
      var off = Math.max(0, Math.min(seekTime, audioBuf.duration));
      var remaining = audioBuf.duration - off;
      _startSource(audioBuf, off, remaining, off,
        "Playing from " + formatTime(off));
    }}

    if (fullAudioBuf) {{
      _play(fullAudioBuf);
    }} else {{
      var buf = b64ToArrayBuffer(fullAudioB64);
      actx.decodeAudioData(buf, function(audioBuf) {{
        fullAudioBuf = audioBuf;
        _play(audioBuf);
      }});
    }}
  }}

  function _startSource(audioBuf, bufOffset, dur, displayOffset, label) {{
    var actx = getAudioCtx();
    var src = actx.createBufferSource();
    src.buffer = audioBuf;
    src.connect(actx.destination);
    currentSource = src;
    playOffset = displayOffset;
    playDuration = dur;
    playStartTime = actx.currentTime;
    pausedAt = -1;
    info.textContent = label;
    setPlaying(true);
    src.onended = function() {{
      currentSource = null;
      if (animId) {{ cancelAnimationFrame(animId); animId = null; }}
      setPlaying(false);
      pausedAt = -1;
      pausedLabel = "";
      info.textContent = "";
      timeSpan.textContent = "";
      drawWaveform();
    }};
    src.start(0, bufOffset, dur);
    animatePlayhead();
  }}

  // --- Hit testing ---

  function getMousePos(e) {{
    var rect = cv.getBoundingClientRect();
    return {{
      x: (e.clientX - rect.left) * (cssW / rect.width),
      y: (e.clientY - rect.top) * (cssH / rect.height)
    }};
  }}

  function hitTestRegion(pos) {{
    if (pos.y < RULER_H) return -1;  // in ruler area
    var t = pos.x / pxPerSec;
    for (var d = 0; d < detections.length; d++) {{
      if (t >= detections[d][0] && t <= detections[d][1]) return d;
    }}
    return -1;
  }}

  function isInRuler(pos) {{
    return pos.y < RULER_H;
  }}

  cv.addEventListener("mousemove", function(e) {{
    var pos = getMousePos(e);
    if (isInRuler(pos)) {{
      // In ruler: show needle, clear region hover
      if (hoveredRegion !== -1) hoveredRegion = -1;
      rulerHoverX = pos.x;
      cv.style.cursor = "pointer";
      drawWaveform();
      return;
    }}
    // In waveform area
    if (rulerHoverX >= 0) {{
      rulerHoverX = -1;
    }}
    var idx = hitTestRegion(pos);
    if (idx !== hoveredRegion) {{
      hoveredRegion = idx;
      cv.style.cursor = idx >= 0 ? "pointer" : "default";
      drawWaveform();
    }}
  }});

  cv.addEventListener("mouseleave", function() {{
    var changed = (hoveredRegion !== -1 || rulerHoverX >= 0);
    hoveredRegion = -1;
    rulerHoverX = -1;
    if (changed) drawWaveform();
  }});

  cv.addEventListener("click", function(e) {{
    var pos = getMousePos(e);
    if (isInRuler(pos)) {{
      // Seek to the clicked time
      var t = Math.max(0, Math.min(pos.x / pxPerSec, duration));
      selectedRegion = -1;
      playFullFrom(t);
      return;
    }}
    var idx = hitTestRegion(pos);
    if (idx >= 0 && regionB64[idx]) {{
      selectedRegion = idx;
      var det = detections[idx];
      var dur = det[1] - det[0];
      playB64(regionB64[idx], det[0], dur,
        "Event " + (idx+1) + ": " + det[0].toFixed(3) + "s \u2013 " + det[1].toFixed(3) + "s");
    }}
  }});

  btnPlay.addEventListener("click", function() {{
    if (playing) {{
      // Currently playing -> pause
      pausePlayback();
    }} else if (pausedAt >= 0) {{
      // Paused -> resume from paused position
      var resumeFrom = pausedAt;
      selectedRegion = -1;
      playFullFrom(resumeFrom);
    }} else {{
      // Stopped -> play from beginning
      selectedRegion = -1;
      playFullFrom(0);
    }}
  }});

  btnStop.addEventListener("click", function() {{
    selectedRegion = -1;
    stopPlayback();
    info.textContent = "";
  }});

  // Stop playback when the widget is removed from the DOM (e.g. cell re-run)
  var observer = new MutationObserver(function(mutations) {{
    if (!document.contains(W)) {{
      destroy();
      observer.disconnect();
      window.removeEventListener("resize", onResize);
    }}
  }});
  observer.observe(document.body, {{childList: true, subtree: true}});

  // Redraw on window resize to fill the new width
  function onResize() {{
    sizeCanvas();
    drawWaveform();
  }}
  window.addEventListener("resize", onResize);

  drawWaveform();
}})();
</script>
"""


def _in_notebook():
    """Return True if running inside a Jupyter notebook kernel."""
    try:
        shell = get_ipython().__class__.__name__  # noqa: F821
        return shell == "ZMQInteractiveShell"
    except NameError:
        return False


def _audio_to_wav_b64(data, sr, sw, ch):
    """Encode raw PCM bytes as a base64 WAV string."""
    buf = io.BytesIO()
    with wave.open(buf, "wb") as wf:
        wf.setframerate(sr)
        wf.setsampwidth(sw)
        wf.setnchannels(ch)
        wf.writeframes(data)
    return base64.b64encode(buf.getvalue()).decode("ascii")


def _downsample_peaks(samples, n_bins):
    """Compute (min, max) peaks per bin for waveform rendering.

    Parameters
    ----------
    samples : 1-D numpy array (one channel, normalised to [-1, 1])
    n_bins : int
        Number of horizontal pixels / bins.

    Returns
    -------
    list of [min, max] pairs, length *n_bins*.
    """
    n = len(samples)
    if n <= n_bins:
        return [[float(s), float(s)] for s in samples]
    bin_size = n / n_bins
    peaks = []
    for i in range(n_bins):
        start = int(i * bin_size)
        end = int((i + 1) * bin_size)
        chunk = samples[start:end]
        peaks.append([float(chunk.min()), float(chunk.max())])
    return peaks


def display_interactive(
    audio_region, regions, energy_threshold=None, height=None
):
    """Build and display an interactive waveform widget in a Jupyter notebook.

    The widget stretches to fill the full cell output width and resizes
    automatically when the browser window is resized.

    Parameters
    ----------
    audio_region : AudioRegion
        The full audio signal.
    regions : list of AudioRegion
        Detected events (each must have ``start`` and ``end`` set).
    energy_threshold : float, optional
        Currently unused (reserved for future threshold line overlay).
    height : int, optional
        Canvas height in pixels.  Defaults to 150 per channel, capped at 400.
    """
    from IPython.display import HTML, display

    sr = audio_region.sr
    sw = audio_region.sw
    ch = audio_region.ch
    duration = audio_region.duration

    if height is None:
        height = min(150 * ch, 400)
    height += _RULER_H  # add space for the time ruler

    # Compute waveform peaks per channel — use 2000 bins (redrawn to actual
    # canvas width by JS, so this just controls the data resolution)
    y = np.asarray(audio_region)
    if y.ndim == 1:
        y = y.reshape(1, -1)

    n_bins = min(2000, y.shape[1])
    peaks = []
    # Compute normalised threshold per channel (same space as peaks)
    if energy_threshold is not None:
        eth_log10 = energy_threshold * np.log(10) / 10
        amplitude_threshold = np.sqrt(np.exp(eth_log10))
    else:
        amplitude_threshold = None
    eth_normalized = []
    for c in range(y.shape[0]):
        channel = y[c].astype(float)
        mx = np.abs(channel).max()
        if mx > 0:
            channel = channel / mx
            if amplitude_threshold is not None:
                eth_normalized.append(min(float(amplitude_threshold / mx), 1.0))
            else:
                eth_normalized.append(None)
        else:
            eth_normalized.append(None)
        peaks.append(_downsample_peaks(channel, n_bins))

    # Detection time ranges
    detections = [[r.start, r.end] for r in regions]

    # Encode each region as a WAV for click-to-play
    regions_b64 = [_audio_to_wav_b64(r.data, sr, sw, ch) for r in regions]

    # Encode the full audio
    full_audio_b64 = _audio_to_wav_b64(audio_region.data, sr, sw, ch)

    wid = "auditok_" + uuid.uuid4().hex[:10]
    title = "{:.2f}s &middot; {} Hz &middot; {}&#x2011;bit &middot; {} ch &middot; {} events".format(
        duration, sr, sw * 8, ch, len(regions)
    )

    html = _WIDGET_HTML.format(
        wid=wid,
        height=height,
        ruler_h=_RULER_H,
        sr=sr,
        nch=ch,
        duration=duration,
        peaks_json=json.dumps(peaks),
        detections_json=json.dumps(detections),
        regions_b64_json=json.dumps(regions_b64),
        eth_json=json.dumps(eth_normalized),
        full_audio_b64=full_audio_b64,
        title=title,
        bg="#482a36",
        plot_bg="#282a36",
        signal_color="#40d970",
    )
    display(HTML(html))
